# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmdet.models.dense_heads import PAAHead
from mmdet.models.task_modules.coders import (
    DeltaXYWHBBoxCoder,
    DistancePointBBoxCoder,
    TBLRBBoxCoder,
)
from mmdet.structures.bbox import BaseBoxes, get_box_tensor
from mmdet.structures.bbox.transforms import distance2bbox
from mmengine import ConfigDict
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import (
    gather_topk,
    get_post_processing_params,
    pad_with_value_if_necessary,
)
from juxtapose.mmdeploy.codebase.mmdet.ops import ncnn_detection_output_forward
from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import Backend, is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.base_dense_head."
    "BaseDenseHead.predict_by_feat"
)
def base_dense_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
    **kwargs,
):
    """Rewrite `predict_by_feat` of `BaseDenseHead` for default backend.
    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.
    Args:
        ctx (ContextCaller): The context with additional information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        score_factors (list[Tensor], optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Defaults to None.
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.
    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor, Tensor]: batch_mlvl_bboxes,
                batch_mlvl_scores, batch_mlvl_centerness
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    num_levels = len(cls_scores)

    featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
    )

    # anchor could be subclass of BaseBoxes in mmrotate
    prior_type = type(mlvl_priors[0])
    mlvl_priors = [get_box_tensor(priors) for priors in mlvl_priors]

    mlvl_priors = [priors.unsqueeze(0) for priors in mlvl_priors]

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]

    if score_factors is None:
        with_score_factors = False
        mlvl_score_factor = [None for _ in range(num_levels)]
    else:
        with_score_factors = True
        mlvl_score_factor = [score_factors[i].detach() for i in range(num_levels)]
        mlvl_score_factors = []
    assert batch_img_metas is not None
    img_shape = batch_img_metas[0]["img_shape"]

    assert len(cls_scores) == len(bbox_preds) == len(mlvl_priors)
    batch_size = cls_scores[0].shape[0]
    cfg = self.test_cfg
    pre_topk = cfg.get("nms_pre", -1)

    mlvl_valid_bboxes = []
    mlvl_valid_scores = []
    mlvl_valid_priors = []

    for cls_score, bbox_pred, score_factors, priors in zip(
        mlvl_cls_scores, mlvl_bbox_preds, mlvl_score_factor, mlvl_priors
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]

        scores = cls_score.permute(0, 2, 3, 1).reshape(
            batch_size, -1, self.cls_out_channels
        )
        if self.use_sigmoid_cls:
            scores = scores.sigmoid()
        else:
            scores = scores.softmax(-1)[:, :, :-1]
        if with_score_factors:
            score_factors = (
                score_factors.permute(0, 2, 3, 1).reshape(batch_size, -1).sigmoid()
            )
            score_factors = score_factors.unsqueeze(2)
        dim = self.bbox_coder.encode_size
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, dim)
        if not is_dynamic_flag:
            priors = priors.data

        if pre_topk > 0:
            priors = pad_with_value_if_necessary(priors, 1, pre_topk)
            bbox_pred = pad_with_value_if_necessary(bbox_pred, 1, pre_topk)
            scores = pad_with_value_if_necessary(scores, 1, pre_topk, 0.0)
            if with_score_factors:
                score_factors = pad_with_value_if_necessary(
                    score_factors, 1, pre_topk, 0.0
                )

            nms_pre_score = scores
            if with_score_factors:
                nms_pre_score = nms_pre_score * score_factors
                if isinstance(self, PAAHead):
                    nms_pre_score = nms_pre_score.sqrt()

            # Get maximum scores for foreground classes.
            if self.use_sigmoid_cls:
                max_scores, _ = nms_pre_score.max(-1)
            else:
                max_scores, _ = nms_pre_score[..., :-1].max(-1)
            _, topk_inds = max_scores.topk(pre_topk)
            bbox_pred, scores, score_factors = gather_topk(
                bbox_pred,
                scores,
                score_factors,
                inds=topk_inds,
                batch_size=batch_size,
                is_batched=True,
            )
            priors = gather_topk(
                priors, inds=topk_inds, batch_size=batch_size, is_batched=False
            )

        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_valid_scores.append(scores)
        mlvl_valid_priors.append(priors)
        if with_score_factors:
            mlvl_score_factors.append(score_factors)

    batch_mlvl_bboxes_pred = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_scores = torch.cat(mlvl_valid_scores, dim=1)
    batch_priors = torch.cat(mlvl_valid_priors, dim=1)

    if issubclass(prior_type, BaseBoxes):
        batch_priors = prior_type(batch_priors, clone=False)

    batch_bboxes = self.bbox_coder.decode(
        batch_priors, batch_mlvl_bboxes_pred, max_shape=img_shape
    )

    batch_bboxes = get_box_tensor(batch_bboxes)

    if with_score_factors:
        batch_score_factors = torch.cat(mlvl_score_factors, dim=1)
    if not self.use_sigmoid_cls:
        batch_scores = batch_scores[..., : self.num_classes]

    if with_score_factors:
        batch_scores = batch_scores * batch_score_factors
        if isinstance(self, PAAHead):
            batch_scores = batch_scores.sqrt()

    if not with_nms:
        return batch_bboxes, batch_scores

    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        batch_bboxes,
        batch_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.base_dense_head."
    "BaseDenseHead.predict_by_feat",
    backend=Backend.RKNN.value,
)
def base_dense_head__predict_by_feat__rknn(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
    **kwargs,
):
    """Rewrite `predict_by_feat` of `BaseDenseHead` for default backend.
    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.
    Args:
        ctx (ContextCaller): The context with additional information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        score_factors (list[Tensor], optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Defaults to None.
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.
    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor, Tensor]: batch_mlvl_bboxes,
                batch_mlvl_scores, batch_mlvl_centerness
    """
    ctx = FUNCTION_REWRITER.get_context()

    # mark nodes for partition
    @mark("BaseDenseHead", outputs=["BaseDenseHead.cls", "BaseDenseHead.loc"])
    def __mark_dense_head(cls_scores, bbox_preds):
        return cls_scores, bbox_preds

    cls_scores, bbox_preds = __mark_dense_head(cls_scores, bbox_preds)

    deploy_cfg = ctx.cfg
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    num_levels = len(cls_scores)

    featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
    )
    mlvl_priors = [priors.unsqueeze(0) for priors in mlvl_priors]

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]
    if score_factors is None:
        with_score_factors = False
        mlvl_score_factor = [None for _ in range(num_levels)]
    else:
        with_score_factors = True
        mlvl_score_factor = [score_factors[i].detach() for i in range(num_levels)]
        mlvl_score_factors = []
    assert batch_img_metas is not None
    img_shape = batch_img_metas[0]["img_shape"]

    assert len(cls_scores) == len(bbox_preds) == len(mlvl_priors)
    batch_size = cls_scores[0].shape[0]

    mlvl_valid_bboxes = []
    mlvl_valid_scores = []
    mlvl_valid_priors = []

    for cls_score, bbox_pred, score_factors, priors in zip(
        mlvl_cls_scores, mlvl_bbox_preds, mlvl_score_factor, mlvl_priors
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]

        scores = cls_score.permute(0, 2, 3, 1).reshape(
            batch_size, -1, self.cls_out_channels
        )
        if self.use_sigmoid_cls:
            scores = scores.sigmoid()
        else:
            scores = scores.softmax(-1)[:, :, :-1]
        if with_score_factors:
            score_factors = (
                score_factors.permute(0, 2, 3, 1).reshape(batch_size, -1).sigmoid()
            )
            score_factors = score_factors.unsqueeze(2)
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4)
        if not is_dynamic_flag:
            priors = priors.data

        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_valid_scores.append(scores)
        mlvl_valid_priors.append(priors)
        if with_score_factors:
            mlvl_score_factors.append(score_factors)

    batch_mlvl_bboxes_pred = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_scores = torch.cat(mlvl_valid_scores, dim=1)
    batch_priors = torch.cat(mlvl_valid_priors, dim=1)
    batch_bboxes = self.bbox_coder.decode(
        batch_priors, batch_mlvl_bboxes_pred, max_shape=img_shape
    )
    if with_score_factors:
        batch_score_factors = torch.cat(mlvl_score_factors, dim=1)
    if not self.use_sigmoid_cls:
        batch_scores = batch_scores[..., : self.num_classes]

    if with_score_factors:
        batch_scores = batch_scores * batch_score_factors
        if isinstance(self, PAAHead):
            batch_scores = batch_scores.sqrt()
    return batch_bboxes, batch_scores


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.base_dense_head."
    "BaseDenseHead.predict_by_feat",
    backend=Backend.NCNN.value,
)
def base_dense_head__predict_by_feat__ncnn(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
    **kwargs,
):
    """Rewrite `predict_by_feat` of BaseDenseHead for ncnn backend.

    Shape node and batch inference is not supported by ncnn. This function
    transform dynamic shape to constant shape and remove batch inference.

    Args:
        ctx (ContextCaller): The context with additional information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        score_factors (list[Tensor], optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Defaults to None.
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.

    Returns:
        output__ncnn (Tensor): outputs, shape is [N, num_det, 6].
    """
    ctx = FUNCTION_REWRITER.get_context()
    assert len(cls_scores) == len(bbox_preds)
    deploy_cfg = ctx.cfg
    assert not is_dynamic_shape(
        deploy_cfg
    ), "base_dense_head for ncnn\
        only supports static shape."

    if score_factors is None:
        # e.g. Retina, FreeAnchor, Foveabox, etc.
        with_score_factors = False
    else:
        # e.g. FCOS, PAA, ATSS, AutoAssign, etc.
        with_score_factors = True
        assert len(cls_scores) == len(score_factors)
    batch_size = cls_scores[0].shape[0]
    assert (
        batch_size == 1
    ), f"ncnn deployment requires batch size 1, \
        got {batch_size}."

    num_levels = len(cls_scores)
    if with_score_factors:
        score_factor_list = score_factors
    else:
        score_factor_list = [None for _ in range(num_levels)]

    if isinstance(self.bbox_coder, DeltaXYWHBBoxCoder):
        vars = torch.tensor(self.bbox_coder.stds)
    elif isinstance(self.bbox_coder, TBLRBBoxCoder):
        normalizer = self.bbox_coder.normalizer
        if isinstance(normalizer, float):
            vars = torch.tensor([normalizer, normalizer, 1, 1], dtype=torch.float32)
        else:
            assert (
                len(normalizer) == 4
            ), f"normalizer of tblr must be 4,\
                got {len(normalizer)}"

            assert (
                normalizer[0] == normalizer[1] and normalizer[2] == normalizer[3]
            ), "normalizer between top \
                        and bottom, left and right must be the same value, or \
                        we can not transform it to delta_xywh format."

            vars = torch.tensor(
                [normalizer[0], normalizer[2], 1, 1], dtype=torch.float32
            )
    elif isinstance(self.bbox_coder, DistancePointBBoxCoder):
        vars = torch.tensor([0, 0, 0, 0], dtype=torch.float32)
    else:
        vars = torch.tensor([1, 1, 1, 1], dtype=torch.float32)

    if isinstance(batch_img_metas[0]["img_shape"][0], int):
        assert isinstance(batch_img_metas[0]["img_shape"][1], int)
        img_height = batch_img_metas[0]["img_shape"][0]
        img_width = batch_img_metas[0]["img_shape"][1]
    else:
        img_height = batch_img_metas[0]["img_shape"][0].item()
        img_width = batch_img_metas[0]["img_shape"][1].item()
    featmap_sizes = [cls_scores[i].shape[-2:] for i in range(num_levels)]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, device=cls_scores[0].device
    )
    batch_mlvl_priors = []
    for i in range(num_levels):
        _priors = mlvl_priors[i].reshape(1, -1, mlvl_priors[i].shape[-1])
        x1 = _priors[:, :, 0:1] / img_width
        y1 = _priors[:, :, 1:2] / img_height
        x2 = _priors[:, :, 2:3] / img_width
        y2 = _priors[:, :, 3:4] / img_height
        priors = torch.cat([x1, y1, x2, y2], dim=2).data
        batch_mlvl_priors.append(priors)

    cfg = self.test_cfg if cfg is None else cfg

    batch_mlvl_bboxes = []
    batch_mlvl_scores = []
    batch_mlvl_score_factors = []

    for level_idx, (cls_score, bbox_pred, score_factor, priors) in enumerate(
        zip(cls_scores, bbox_preds, score_factor_list, batch_mlvl_priors)
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]
        # ncnn needs 3 dimensions to reshape when including -1 parameter in
        # width or height dimension.
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4)
        if with_score_factors:
            score_factor = (
                score_factor.permute(0, 2, 3, 1).reshape(batch_size, -1, 1).sigmoid()
            )
        cls_score = cls_score.permute(0, 2, 3, 1).reshape(
            batch_size, -1, self.cls_out_channels
        )
        # ncnn DetectionOutput op needs num_class + 1 classes. So if sigmoid
        # score, we should padding background class according to mmdetection
        # num_class definition.
        if self.use_sigmoid_cls:
            scores = cls_score.sigmoid()
            dummy_background_score = torch.zeros(
                batch_size, cls_score.shape[1], 1, device=cls_score.device
            )
            scores = torch.cat([scores, dummy_background_score], dim=2)
        else:
            scores = cls_score.softmax(-1)
        batch_mlvl_bboxes.append(bbox_pred)
        batch_mlvl_scores.append(scores)
        batch_mlvl_score_factors.append(score_factor)

    batch_mlvl_priors = torch.cat(batch_mlvl_priors, dim=1)
    batch_mlvl_scores = torch.cat(batch_mlvl_scores, dim=1)
    batch_mlvl_bboxes = torch.cat(batch_mlvl_bboxes, dim=1)
    batch_mlvl_scores = torch.cat(
        [
            batch_mlvl_scores[:, :, self.num_classes :],
            batch_mlvl_scores[:, :, 0 : self.num_classes],
        ],
        dim=2,
    )
    if isinstance(self.bbox_coder, TBLRBBoxCoder):
        batch_mlvl_bboxes = _tblr_pred_to_delta_xywh_pred(batch_mlvl_bboxes, vars[0:2])
    elif isinstance(self.bbox_coder, DistancePointBBoxCoder):
        bboxes_x0 = batch_mlvl_bboxes[:, :, 0:1] / img_width
        bboxes_y0 = batch_mlvl_bboxes[:, :, 1:2] / img_height
        bboxes_x1 = batch_mlvl_bboxes[:, :, 2:3] / img_width
        bboxes_y1 = batch_mlvl_bboxes[:, :, 3:4] / img_height
        batch_mlvl_bboxes = torch.cat(
            [bboxes_x0, bboxes_y0, bboxes_x1, bboxes_y1], dim=2
        )
        batch_mlvl_priors = distance2bbox(batch_mlvl_priors, batch_mlvl_bboxes)

    if with_score_factors:
        batch_mlvl_score_factors = torch.cat(batch_mlvl_score_factors, dim=1)
        batch_mlvl_scores = batch_mlvl_scores.permute(0, 2, 1).unsqueeze(
            3
        ) * batch_mlvl_score_factors.permute(0, 2, 1).unsqueeze(3)
        batch_mlvl_scores = batch_mlvl_scores.squeeze(3).permute(0, 2, 1)
        if isinstance(self, PAAHead):
            batch_mlvl_scores = batch_mlvl_scores.sqrt()

    # flatten for ncnn DetectionOutput op inputs.
    batch_mlvl_vars = vars.expand_as(batch_mlvl_priors)
    batch_mlvl_bboxes = batch_mlvl_bboxes.reshape(batch_size, 1, -1)
    batch_mlvl_scores = batch_mlvl_scores.reshape(batch_size, 1, -1)
    batch_mlvl_priors = batch_mlvl_priors.reshape(batch_size, 1, -1)
    batch_mlvl_vars = batch_mlvl_vars.reshape(batch_size, 1, -1)
    batch_mlvl_priors = torch.cat([batch_mlvl_priors, batch_mlvl_vars], dim=1)
    if not isinstance(self.bbox_coder, DistancePointBBoxCoder):
        batch_mlvl_priors = batch_mlvl_priors.data

    post_params = get_post_processing_params(ctx.cfg)
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)

    output__ncnn = ncnn_detection_output_forward(
        batch_mlvl_bboxes,
        batch_mlvl_scores,
        batch_mlvl_priors,
        score_threshold,
        iou_threshold,
        pre_top_k,
        keep_top_k,
        self.num_classes + 1,
        vars.cpu().detach().numpy(),
    )

    return output__ncnn


def _tblr_pred_to_delta_xywh_pred(
    bbox_pred: torch.Tensor, normalizer: torch.Tensor
) -> torch.Tensor:
    """Transform tblr format bbox prediction to delta_xywh format for ncnn.

    An internal function for transforming tblr format bbox prediction to
    delta_xywh format. ncnn DetectionOutput layer needs delta_xywh format
    bbox_pred as input.

    Args:
        bbox_pred (Tensor): The bbox prediction of tblr format, has shape
            (N, num_det, 4).
        normalizer (Tensor): The normalizer scale of bbox horizon and
            vertical coordinates, has shape (2,).

    Returns:
        Tensor: The delta_xywh format bbox predictions.
    """
    top = bbox_pred[:, :, 0:1]
    bottom = bbox_pred[:, :, 1:2]
    left = bbox_pred[:, :, 2:3]
    right = bbox_pred[:, :, 3:4]
    h = (top + bottom) * normalizer[0]
    w = (left + right) * normalizer[1]

    _dwh = torch.cat([w, h], dim=2)
    assert torch.all(_dwh >= 0), "wh must be positive before log."
    dwh = torch.log(_dwh)

    return torch.cat([(right - left) / 2, (bottom - top) / 2, dwh], dim=2)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

import torch
from torch.onnx import symbolic_helper

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


class GridPriorsTRTOp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, base_anchors, feat_h, feat_w, stride_h: int, stride_w: int):
        """Generate grid priors by base anchors."""

        # torch>=1.13 has runtime error
        # when using torch.arange in autograd function
        output = getattr(GridPriorsTRTOp, "output", None)
        if output is not None:
            return output

        device = base_anchors.device
        dtype = base_anchors.dtype
        shift_x = torch.arange(0, feat_w, device=device).to(dtype) * stride_w
        shift_y = torch.arange(0, feat_h, device=device).to(dtype) * stride_h

        def _meshgrid(x, y, row_major=True):
            # use shape instead of len to keep tracing while exporting to onnx
            xx = x.repeat(y.shape[0])
            yy = y.view(-1, 1).repeat(1, x.shape[0]).view(-1)
            if row_major:
                return xx, yy
            else:
                return yy, xx

        shift_xx, shift_yy = _meshgrid(shift_x, shift_y)
        shifts = torch.stack([shift_xx, shift_yy, shift_xx, shift_yy], dim=-1)

        all_anchors = base_anchors[None, :, :] + shifts[:, None, :]
        all_anchors = all_anchors.view(-1, 4)
        # then (0, 1), (0, 2), ...
        return all_anchors

    @staticmethod
    @symbolic_helper.parse_args("v", "v", "v", "i", "i")
    def symbolic(g, base_anchors, feat_h, feat_w, stride_h: int, stride_w: int):
        """Map ops to onnx symbolics."""
        # zero_h and zero_w is used to provide shape to GridPriorsTRT
        feat_h = symbolic_helper._unsqueeze_helper(g, feat_h, [0])
        feat_w = symbolic_helper._unsqueeze_helper(g, feat_w, [0])
        zero_h = g.op(
            "ConstantOfShape",
            feat_h,
            value_t=torch.tensor([0], dtype=torch.long),
        )
        zero_w = g.op(
            "ConstantOfShape",
            feat_w,
            value_t=torch.tensor([0], dtype=torch.long),
        )
        return g.op(
            "mmdeploy::GridPriorsTRT",
            base_anchors,
            zero_h,
            zero_w,
            stride_h_i=stride_h,
            stride_w_i=stride_w,
        )


grid_priors_trt = GridPriorsTRTOp.apply


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.prior_generators."
    "AnchorGenerator.single_level_grid_priors",
    backend="tensorrt",
)
def anchorgenerator__single_level_grid_priors__trt(
    self,
    featmap_size: Tuple[int],
    level_idx: int,
    dtype: torch.dtype = torch.float32,
    device: str = "cuda",
) -> torch.Tensor:
    """This is a rewrite to replace ONNX anchor generator to TensorRT custom
    op.

    Args:
        ctx : The rewriter context
        featmap_size (tuple[int]): Size of the feature maps.
        level_idx (int): The index of corresponding feature map level.
        dtype (obj:`torch.dtype`): Date type of points.Defaults to
            ``torch.float32``.
        device (str, optional): The device the tensor will be put on.
            Defaults to 'cuda'.

        Returns:
            torch.Tensor: Anchors in the overall feature maps.
    """
    ctx = FUNCTION_REWRITER.get_context()
    from mmdet.models.task_modules.prior_generators import AnchorGenerator

    if type(self) != AnchorGenerator:
        # only use custom node on default generator.
        return ctx.origin_func(
            self,
            featmap_size=featmap_size,
            level_idx=level_idx,
            dtype=dtype,
            device=device,
        )
    feat_h, feat_w = featmap_size
    output = ctx.origin_func(self, featmap_size, level_idx, dtype, device).data
    if isinstance(feat_h, int) and isinstance(feat_w, int):
        return output
    base_anchors = self.base_anchors[level_idx].to(device).to(dtype)
    stride_w, stride_h = self.strides[level_idx]

    GridPriorsTRTOp.output = output
    return grid_priors_trt(base_anchors, feat_h, feat_w, stride_h, stride_w)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.triu")
@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.triu")
def triu__default(
    input: torch.Tensor, diagonal: int = 0, *args, **kwargs
) -> torch.Tensor:
    """Rewrite `triu` for exporting model to ONNX."""
    assert len(input.shape) >= 2
    height, width = input.shape[-2:]
    arange0 = torch.arange(width, device=input.device).unsqueeze(0)
    arange1 = torch.arange(height, device=input.device).unsqueeze(-1)
    mask = arange0 >= torch.add(arange1, diagonal)
    return input * mask



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Dict, Optional, Sequence

import numpy as np
import onnxruntime as ort
import torch

from juxtapose.mmdeploy.utils import Backend, get_root_logger, parse_device_id
from juxtapose.mmdeploy.utils.timer import TimeCounter
from ..base import BACKEND_WRAPPER, BaseWrapper
from .init_plugins import get_ops_path


@BACKEND_WRAPPER.register_module(Backend.ONNXRUNTIME.value)
class ORTWrapper(BaseWrapper):
    """ONNXRuntime wrapper for inference.

    Args:
        onnx_file (str): Input onnx model file.
        device (str): The device to input model.
        output_names (Sequence[str] | None): Names of model outputs in order.
           Defaults to `None` and the wrapper will load the output names from
           model.

    Examples:
        >>> from juxtapose.mmdeploy.backend.onnxruntime import ORTWrapper
        >>> import torch
        >>>
        >>> onnx_file = 'model.onnx'
        >>> model = ORTWrapper(onnx_file, 'cpu')
        >>> inputs = dict(input=torch.randn(1, 3, 224, 224, device='cpu'))
        >>> outputs = model(inputs)
        >>> print(outputs)
    """

    def __init__(
        self, onnx_file: str, device: str, output_names: Optional[Sequence[str]] = None
    ):
        # get the custom op path
        ort_custom_op_path = get_ops_path()
        session_options = ort.SessionOptions()
        # register custom op for onnxruntime
        logger = get_root_logger()
        if osp.exists(ort_custom_op_path):
            session_options.register_custom_ops_library(ort_custom_op_path)
            logger.info(
                "Successfully loaded onnxruntime custom ops from "
                f"{ort_custom_op_path}"
            )
        else:
            logger.warning(
                "The library of onnxruntime custom ops does"
                f"not exist: {ort_custom_op_path}"
            )
        device_id = parse_device_id(device)
        providers = (
            ["CPUExecutionProvider"]
            if device == "cpu"
            else [("CUDAExecutionProvider", {"device_id": device_id})]
        )
        sess = ort.InferenceSession(onnx_file, session_options, providers=providers)
        if output_names is None:
            output_names = [_.name for _ in sess.get_outputs()]
        self.sess = sess
        self._input_metas = {_.name: _ for _ in sess.get_inputs()}
        self.io_binding = sess.io_binding()
        self.device_id = device_id
        self.device_type = "cpu" if device == "cpu" else "cuda"
        super().__init__(output_names)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Run forward inference.

        Args:
            inputs (Dict[str, torch.Tensor]): The input name and tensor pairs.

        Returns:
            Dict[str, torch.Tensor]: The output name and tensor pairs.
        """
        for name, input_tensor in inputs.items():
            # set io binding for inputs/outputs
            input_type = self._input_metas[name].type
            if "float16" in input_type:
                input_tensor = input_tensor.to(torch.float16)
            input_tensor = input_tensor.contiguous()
            if self.device_type == "cpu":
                input_tensor = input_tensor.cpu()
            # Avoid unnecessary data transfer between host and device
            element_type = input_tensor.new_zeros(1, device="cpu").numpy().dtype
            self.io_binding.bind_input(
                name=name,
                device_type=self.device_type,
                device_id=self.device_id,
                element_type=element_type,
                shape=input_tensor.shape,
                buffer_ptr=input_tensor.data_ptr(),
            )

        for name in self._output_names:
            self.io_binding.bind_output(name)
        # run session to get outputs
        if self.device_type == "cuda":
            torch.cuda.synchronize()
        self.__ort_execute(self.io_binding)
        output_list = self.io_binding.copy_outputs_to_cpu()
        outputs = {}
        for output_name, numpy_tensor in zip(self._output_names, output_list):
            if numpy_tensor.dtype == np.float16:
                numpy_tensor = numpy_tensor.astype(np.float32)
            outputs[output_name] = torch.from_numpy(numpy_tensor)

        return outputs

    @TimeCounter.count_time(Backend.ONNXRUNTIME.value)
    def __ort_execute(self, io_binding: ort.IOBinding):
        """Run inference with ONNXRuntime session.

        Args:
            io_binding (ort.IOBinding): To bind input/output to a specified
                device, e.g. GPU.
        """
        self.sess.run_with_iobinding(io_binding)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.layers.matrix_nms.mask_matrix_nms"
)
def mask_matrix_nms__default(
    masks,
    labels,
    scores,
    filter_thr=-1,
    nms_pre=-1,
    max_num=-1,
    kernel="gaussian",
    sigma=2.0,
    mask_area=None,
):
    """Matrix NMS for multi-class masks.

    Args:
        masks (Tensor): Has shape (num_instances, h, w)
        labels (Tensor): Labels of corresponding masks,
            has shape (num_instances,).
        scores (Tensor): Mask scores of corresponding masks,
            has shape (num_instances).
        filter_thr (float): Score threshold to filter the masks
            after matrix nms. Default: -1, which means do not
            use filter_thr.
        nms_pre (int): The max number of instances to do the matrix nms.
            Default: -1, which means do not use nms_pre.
        max_num (int, optional): If there are more than max_num masks after
            matrix, only top max_num will be kept. Default: -1, which means
            do not use max_num.
        kernel (str): 'linear' or 'gaussian'.
        sigma (float): std in gaussian method.
        mask_area (Tensor): The sum of seg_masks.

    Returns:
        tuple(Tensor): Processed mask results.

            - scores (Tensor): Updated scores, has shape (n,).
            - labels (Tensor): Remained labels, has shape (n,).
            - masks (Tensor): Remained masks, has shape (n, w, h).
            - keep_inds (Tensor): The indices number of
                the remaining mask in the input mask, has shape (n,).
    """
    assert len(labels) == len(masks) == len(scores)
    assert len(masks) == len(mask_area)
    # sort and keep top nms_pre
    nms_pre = max(0, nms_pre)
    if nms_pre <= 0:
        nms_pre = scores.shape[0]
    scores, sort_inds = torch.topk(scores, nms_pre)

    keep_inds = sort_inds
    masks = masks[sort_inds]
    mask_area = mask_area[sort_inds]
    labels = labels[sort_inds]
    num_masks = labels.size(0)
    flatten_masks = masks.reshape(num_masks, -1).float()
    # inter.
    inter_matrix = torch.mm(flatten_masks, flatten_masks.transpose(1, 0))
    expanded_mask_area = mask_area.unsqueeze(1)
    total_area = expanded_mask_area + expanded_mask_area.transpose(1, 0) - inter_matrix
    total_mask = total_area > 0
    total_area = total_area.where(total_mask, total_area.new_ones(1))
    iou_matrix = (inter_matrix / total_area).triu(diagonal=1)
    expanded_labels = labels.unsqueeze(1)
    label_matrix = expanded_labels == expanded_labels.transpose(1, 0)

    # iou decay
    decay_iou = iou_matrix.where(label_matrix, iou_matrix.new_zeros(1))

    # iou compensation
    compensate_iou, _ = decay_iou.max(0)
    compensate_iou = compensate_iou.expand(num_masks, num_masks).transpose(1, 0)

    # calculate the decay_coefficient
    if kernel == "gaussian":
        decay_matrix = torch.exp(-1 * sigma * (decay_iou**2))
        compensate_matrix = torch.exp(-1 * sigma * (compensate_iou**2))
        decay_coefficient, _ = (decay_matrix / compensate_matrix).min(0)
    elif kernel == "linear":
        decay_matrix = (1 - decay_iou) / (1 - compensate_iou)
        decay_coefficient, _ = decay_matrix.min(0)
    else:
        raise NotImplementedError(f"{kernel} kernel is not supported in matrix nms!")
    # update the score.
    scores = scores * decay_coefficient

    keep = scores >= filter_thr
    scores = scores.where(keep, scores.new_zeros(1))

    # sort and keep top max_num
    scores, sort_inds = torch.topk(scores, max(max_num, 0))
    keep_inds = keep_inds[sort_inds]
    masks = masks[sort_inds]
    labels = labels[sort_inds]

    return scores, labels, masks, keep_inds



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified from:
# https://github.com/pytorch/pytorch/blob/9ade03959392e5a90b74261012de1d806cab2253/torch/onnx/symbolic_opset9.py

import torch
from torch.onnx.symbolic_helper import (
    _get_tensor_dim_size,
    _get_tensor_rank,
    _unimplemented,
    _unsqueeze_helper,
    parse_args,
)

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@parse_args("v", "i", "v", "v", "f", "i")
def instance_norm(g, input, num_groups, weight, bias, eps, cudnn_enabled):
    """Symbolic function for `instance_norm`."""
    channel_size = _get_tensor_dim_size(input, 1)
    if channel_size is not None:
        assert channel_size % num_groups == 0
    input_rank = _get_tensor_rank(input)
    if input_rank is None:
        return _unimplemented("group_norm", "unknown input rank")
    # 0 in the shape list keeps dimension value unchanged.
    shape = [0, num_groups, -1]
    input_reshaped = g.op(
        "Reshape", input, g.op("Constant", value_t=torch.LongTensor(shape))
    )

    # C is always divisible by num_groups
    # Due to shape difference. we need to apply weight and bias after
    # instance norm computation and reshape
    weight_ = g.op(
        "Constant",
        value_t=torch.tensor([1.0] * num_groups).type(
            "torch." + input.type().scalarType() + "Tensor"
        ),
    )
    bias_ = g.op(
        "Constant",
        value_t=torch.tensor([0.0] * num_groups).type(
            "torch." + input.type().scalarType() + "Tensor"
        ),
    )

    norm_reshaped = g.op(
        "mmdeploy::TRTInstanceNormalization",
        input_reshaped,
        weight_,
        bias_,
        epsilon_f=eps,
    )
    norm = g.op("Reshape", norm_reshaped, g.op("Shape", input))

    if weight is None or weight.node().mustBeNone():
        weight_value = torch.tensor([1.0]).type(
            "torch." + input.type().scalarType() + "Tensor"
        )
        weight = g.op("Constant", value_t=weight_value)
    if bias is None or bias.node().mustBeNone():
        bias_value = torch.tensor([0.0]).type(
            "torch." + input.type().scalarType() + "Tensor"
        )
        bias = g.op("Constant", value_t=bias_value)

    # Norm has shape [N, C, *] so we reshape weight and bias to [C, *]
    axes = list(range(1, input_rank - 1))
    from torch.onnx.symbolic_opset9 import add, mul

    return add(
        g,
        mul(g, norm, _unsqueeze_helper(g, weight, axes)),
        _unsqueeze_helper(g, bias, axes),
    )


@SYMBOLIC_REWRITER.register_symbolic("group_norm", backend="tensorrt", is_pytorch=True)
def instance_norm__tensorrt(*args):
    """Register symbolic function for TensorRT backend.

    Notes:
        Instance normalization is implemented in group norm in pytorch.
    """
    return instance_norm(*args)



# Copyright (c) OpenMMLab. All rights reserved.

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@SYMBOLIC_REWRITER.register_symbolic(
    "adaptive_avg_pool2d", is_pytorch=True, backend="ncnn"
)
def adaptive_avg_pool2d__ncnn(g, x, output_size):
    """Register ncnn symbolic function for `adaptive_avg_pool2d`.

    Align symbolic of adaptive_avg_pool2d in ncnn.
    """
    return g.op("mmdeploy::AdaptiveAvgPool2d", x, output_size)



# Copyright (c) OpenMMLab. All rights reserved.
from . import delta_xywh_bbox_coder  # noqa: F401,F403
from . import distance_point_bbox_coder  # noqa: F401,F403
from . import tblr_bbox_coder  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
from .calibration import create_calib_input_data
from .extract_model import extract_model
from .inference import inference_model
from .pytorch2onnx import torch2onnx
from .pytorch2torchscript import torch2torchscript
from .utils import build_task_processor, get_predefined_partition_cfg
from .visualize import visualize_model

__all__ = [
    'create_calib_input_data', 'extract_model', 'inference_model',
    'torch2onnx', 'torch2torchscript', 'build_task_processor',
    'get_predefined_partition_cfg', 'visualize_model'
]



# Copyright (c) OpenMMLab. All rights reserved.
# recovery for mmyolo
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms  # noqa: F401, F403
from . import matrix_nms  # noqa: F401, F403

__all__ = ["multiclass_nms"]



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.tvm import get_library_ext, is_available
from ..core import PIPELINE_MANAGER

__all__ = ["is_available", "get_library_ext"]

if is_available():
    from juxtapose.mmdeploy.backend.tvm import HDF5Dataset
    from juxtapose.mmdeploy.backend.tvm import from_onnx as _from_onnx

    from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)

    __all__ += ["from_onnx", "HDF5Dataset"]



_base_ = [
    'mmdet::_base_/default_runtime.py', 'mmdet::_base_/schedules/schedule_1x.py',
    'mmdet::_base_/datasets/coco_detection.py', 'mmdet::rtmdet/rtmdet_tta.py'
]
model = dict(
    type='RTMDet',
    data_preprocessor=dict(
        type='DetDataPreprocessor',
        mean=[103.53, 116.28, 123.675],
        std=[57.375, 57.12, 58.395],
        bgr_to_rgb=False,
        batch_augments=None),
    backbone=dict(
        type='CSPNeXt',
        arch='P5',
        expand_ratio=0.5,
        deepen_factor=1,
        widen_factor=1,
        channel_attention=True,
        norm_cfg=dict(type='SyncBN'),
        act_cfg=dict(type='SiLU', inplace=True)),
    neck=dict(
        type='CSPNeXtPAFPN',
        in_channels=[256, 512, 1024],
        out_channels=256,
        num_csp_blocks=3,
        expand_ratio=0.5,
        norm_cfg=dict(type='SyncBN'),
        act_cfg=dict(type='SiLU', inplace=True)),
    bbox_head=dict(
        type='RTMDetSepBNHead',
        num_classes=80,
        in_channels=256,
        stacked_convs=2,
        feat_channels=256,
        anchor_generator=dict(
            type='MlvlPointGenerator', offset=0, strides=[8, 16, 32]),
        bbox_coder=dict(type='DistancePointBBoxCoder'),
        loss_cls=dict(
            type='QualityFocalLoss',
            use_sigmoid=True,
            beta=2.0,
            loss_weight=1.0),
        loss_bbox=dict(type='GIoULoss', loss_weight=2.0),
        with_objectness=False,
        exp_on_reg=True,
        share_conv=True,
        pred_kernel_size=1,
        norm_cfg=dict(type='SyncBN'),
        act_cfg=dict(type='SiLU', inplace=True)),
    train_cfg=dict(
        assigner=dict(type='DynamicSoftLabelAssigner', topk=13),
        allowed_border=-1,
        pos_weight=-1,
        debug=False),
    test_cfg=dict(
        nms_pre=30000,
        min_bbox_size=0,
        score_thr=0.001,
        nms=dict(type='nms', iou_threshold=0.65),
        max_per_img=300),
)

train_pipeline = [
    dict(type='LoadImageFromFile', backend_args={{_base_.backend_args}}),
    dict(type='LoadAnnotations', with_bbox=True),
    dict(type='CachedMosaic', img_scale=(640, 640), pad_val=114.0),
    dict(
        type='RandomResize',
        scale=(1280, 1280),
        ratio_range=(0.1, 2.0),
        keep_ratio=True),
    dict(type='RandomCrop', crop_size=(640, 640)),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(type='Pad', size=(640, 640), pad_val=dict(img=(114, 114, 114))),
    dict(
        type='CachedMixUp',
        img_scale=(640, 640),
        ratio_range=(1.0, 1.0),
        max_cached_images=20,
        pad_val=(114, 114, 114)),
    dict(type='mmdet.PackDetInputs')
]

train_pipeline_stage2 = [
    dict(type='LoadImageFromFile', backend_args={{_base_.backend_args}}),
    dict(type='LoadAnnotations', with_bbox=True),
    dict(
        type='RandomResize',
        scale=(640, 640),
        ratio_range=(0.1, 2.0),
        keep_ratio=True),
    dict(type='RandomCrop', crop_size=(640, 640)),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(type='Pad', size=(640, 640), pad_val=dict(img=(114, 114, 114))),
    dict(type='mmdet.PackDetInputs')
]

test_pipeline = [
    dict(type='LoadImageFromFile', backend_args={{_base_.backend_args}}),
    dict(type='Resize', scale=(640, 640), keep_ratio=True),
    dict(type='Pad', size=(640, 640), pad_val=dict(img=(114, 114, 114))),
    dict(
        type='mmdet.PackDetInputs',
        meta_keys=('img_id', 'img_path', 'ori_shape', 'img_shape',
                   'scale_factor'))
]

train_dataloader = dict(
    batch_size=32,
    num_workers=10,
    batch_sampler=None,
    pin_memory=True,
    dataset=dict(pipeline=train_pipeline))
val_dataloader = dict(
    batch_size=5, num_workers=10, dataset=dict(pipeline=test_pipeline))
test_dataloader = val_dataloader

max_epochs = 300
stage2_num_epochs = 20
base_lr = 0.004
interval = 10

train_cfg = dict(
    max_epochs=max_epochs,
    val_interval=interval,
    dynamic_intervals=[(max_epochs - stage2_num_epochs, 1)])

val_evaluator = dict(proposal_nums=(100, 1, 10))
test_evaluator = val_evaluator

# optimizer
optim_wrapper = dict(
    _delete_=True,
    type='OptimWrapper',
    optimizer=dict(type='AdamW', lr=base_lr, weight_decay=0.05),
    paramwise_cfg=dict(
        norm_decay_mult=0, bias_decay_mult=0, bypass_duplicate=True))

# learning rate
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0e-5,
        by_epoch=False,
        begin=0,
        end=1000),
    dict(
        # use cosine lr from 150 to 300 epoch
        type='CosineAnnealingLR',
        eta_min=base_lr * 0.05,
        begin=max_epochs // 2,
        end=max_epochs,
        T_max=max_epochs // 2,
        by_epoch=True,
        convert_to_iter_based=True),
]

# hooks
default_hooks = dict(
    checkpoint=dict(
        interval=interval,
        max_keep_ckpts=3  # only keep latest 3 checkpoints
    ))
custom_hooks = [
    dict(
        type='EMAHook',
        ema_type='ExpMomentumEMA',
        momentum=0.0002,
        update_buffers=True,
        priority=49),
    dict(
        type='PipelineSwitchHook',
        switch_epoch=max_epochs - stage2_num_epochs,
        switch_pipeline=train_pipeline_stage2)
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.__getattribute__", backend="ncnn"
)
def tensor__getattribute__ncnn(self: torch.Tensor, name: str):
    """Rewrite `__getattribute__` of `torch.Tensor` for ncnn backend.

    Shape node is not supported by ncnn. This function transform dynamic shape
    to constant shape.
    """

    ctx = FUNCTION_REWRITER.get_context()
    ret = ctx.origin_func(self, name)
    if name == "shape":
        ret = torch.Size([int(s) for s in ret])
    return ret



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

import mmengine
import numpy as np
import torch

from juxtapose.mmdeploy.utils import Backend, get_backend, get_input_shape, load_config


def visualize_model(
    model_cfg: Union[str, mmengine.Config],
    deploy_cfg: Union[str, mmengine.Config],
    model: Union[str, Sequence[str]],
    img: Union[str, np.ndarray, Sequence[str]],
    device: str,
    backend: Optional[Backend] = None,
    output_file: Optional[str] = None,
    show_result: bool = False,
    **kwargs,
):
    """Run inference with PyTorch or backend model and show results.

    Examples:
        >>> from juxtapose.mmdeploy.apis import visualize_model
        >>> model_cfg = ('mmdetection/configs/fcos/'
                         'fcos_r50_caffe_fpn_gn-head_1x_coco.py')
        >>> deploy_cfg = ('configs/mmdet/detection/'
                          'detection_onnxruntime_dynamic.py')
        >>> model = 'work_dir/fcos.onnx'
        >>> img = 'demo.jpg'
        >>> device = 'cpu'
        >>> visualize_model(model_cfg, deploy_cfg, model, \
            img, device, show_result=True)

    Args:
        model_cfg (str | mmengine.Config): Model config file or Config object.
        deploy_cfg (str | mmengine.Config): Deployment config file or Config
            object.
        model (str | Sequence[str]): Input model or file(s).
        img (str | np.ndarray | Sequence[str]): Input image file or numpy array
            for inference.
        device (str): A string specifying device type.
        backend (Backend): Specifying backend type, defaults to `None`.
        output_file (str): Output file to save visualized image, defaults to
            `None`. Only valid if `show_result` is set to `False`.
        show_result (bool): Whether to show plotted image in windows, defaults
            to `False`.
    """
    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

    from juxtapose.mmdeploy.apis.utils import build_task_processor

    task_processor = build_task_processor(model_cfg, deploy_cfg, device)

    input_shape = get_input_shape(deploy_cfg)
    if backend is None:
        backend = get_backend(deploy_cfg)

    if isinstance(model, str):
        model = [model]

    if isinstance(model, (list, tuple)):
        assert len(model) > 0, "Model should have at least one element."
        if backend == Backend.PYTORCH:
            model = task_processor.build_pytorch_model(model[0])
        else:
            model = task_processor.build_backend_model(
                model, data_preprocessor_updater=task_processor.update_data_preprocessor
            )

    model_inputs, _ = task_processor.create_input(img, input_shape)
    with torch.no_grad():
        result = model.test_step(model_inputs)[0]

    if show_result:
        try:
            # check headless
            import tkinter

            tkinter.Tk()
        except Exception as e:
            from juxtapose.mmdeploy.utils import get_root_logger

            logger = get_root_logger()
            logger.warning(
                f"render and display result skipped for headless device, exception {e}"  # noqa: E501
            )
            show_result = False

    if isinstance(img, str) or not isinstance(img, Sequence):
        img = [img]
    for single_img in img:
        task_processor.visualize(
            image=single_img,
            model=model,
            result=result,
            output_file=output_file,
            window_name=backend.value,
            show_result=show_result,
        )



# Copyright (c) OpenMMLab. All rights reserved.
import math
from functools import partial
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from mmengine import Config
from mmengine.model.base_model.data_preprocessor import BaseDataPreprocessor
from mmengine.registry import Registry
from mmengine.structures import BaseDataElement, InstanceData
from torch import Tensor, nn

from juxtapose.mmdeploy.backend.base import get_backend_file_count
from juxtapose.mmdeploy.codebase.base import BaseBackendModel
from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import (
    Backend,
    get_backend,
    get_codebase_config,
    get_ir_config,
    get_partition_config,
    get_quantization_config,
    load_config,
)

# Use registry to store models with different partition methods
# If a model doesn't need to partition, we don't need this registry
__BACKEND_MODEL = Registry("backend_detectors")


@__BACKEND_MODEL.register_module("end2end")
class End2EndModel(BaseBackendModel):
    """End to end model for inference of detection.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string specifying device type.
        deploy_cfg (str|Config): Deployment config file or loaded Config
            object.
        data_preprocessor (dict|nn.Module): The data preprocessor.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        deploy_cfg: Union[str, Config],
        model_cfg: Optional[Union[str, Config]] = None,
        data_preprocessor: Optional[Union[dict, nn.Module]] = None,
        **kwargs,
    ):
        super().__init__(deploy_cfg=deploy_cfg, data_preprocessor=data_preprocessor)
        self.deploy_cfg = deploy_cfg
        self.model_cfg = model_cfg
        self.device = device
        self._init_wrapper(backend=backend, backend_files=backend_files, device=device)

    def _init_wrapper(
        self, backend: Backend, backend_files: Sequence[str], device: str
    ):
        """Initialize backend wrapper.

        Args:
            backend (Backend): The backend enum, specifying backend type.
            backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
            device (str): A string specifying device type.
        """
        output_names = self.output_names
        self.wrapper = BaseBackendModel._build_wrapper(
            backend=backend,
            backend_files=backend_files,
            device=device,
            input_names=[self.input_name],
            output_names=output_names,
            deploy_cfg=self.deploy_cfg,
        )

    @staticmethod
    def __clear_outputs(
        test_outputs: List[Union[Tensor, np.ndarray]]
    ) -> List[Union[List[Tensor], List[np.ndarray]]]:
        """Removes additional outputs and detections with zero and negative
        score.

        Args:
            test_outputs (List[Union[Tensor, np.ndarray]]):
                outputs of forward_test.

        Returns:
            List[Union[List[Tensor], List[np.ndarray]]]:
                outputs with without zero score object.
        """
        batch_size = len(test_outputs[0])

        num_outputs = len(test_outputs)
        outputs = [[None for _ in range(batch_size)] for _ in range(num_outputs)]

        for i in range(batch_size):
            inds = test_outputs[0][i, :, 4] > 0.0
            for output_id in range(num_outputs):
                outputs[output_id][i] = test_outputs[output_id][i, inds, ...]
        return outputs

    @staticmethod
    def postprocessing_masks(
        det_bboxes: Union[np.ndarray, Tensor],
        det_masks: Union[np.ndarray, Tensor],
        img_w: int,
        img_h: int,
        device: str = "cpu",
    ) -> Tensor:
        """Additional processing of masks. Resizes masks from [num_det, 28, 28]
        to [num_det, img_w, img_h]. Analog of the 'mmdeploy.codebase.mmdet.
        models.roi_heads.fcn_mask_head._do_paste_mask' function.

        Args:
            det_bboxes (np.ndarray | Tensor): Bbox of shape [num_det, 4]
            det_masks (np.ndarray | Tensor): Masks of shape [num_det, 28, 28].
            img_w (int): Width of the original image.
            img_h (int): Height of the original image.
            device :(str): The device type.

        Returns:
            Tensor: masks of shape [N, num_det, img_h, img_w].
        """
        masks = det_masks
        bboxes = det_bboxes
        device = torch.device(device)
        num_det = bboxes.shape[0]
        # Skip postprocessing if no detections are found.
        if num_det == 0:
            return torch.zeros(0, img_h, img_w, dtype=torch.float32, device=device)

        if isinstance(masks, np.ndarray):
            masks = torch.tensor(masks, device=device)
            bboxes = torch.tensor(bboxes, device=device)

        masks = masks.to(device)
        bboxes = bboxes.to(device)

        result_masks = []
        for bbox, mask in zip(bboxes, masks):
            x0_int, y0_int = 0, 0
            x1_int, y1_int = img_w, img_h

            img_y = (
                torch.arange(y0_int, y1_int, dtype=torch.float32, device=device) + 0.5
            )
            img_x = (
                torch.arange(x0_int, x1_int, dtype=torch.float32, device=device) + 0.5
            )
            x0, y0, x1, y1 = bbox

            img_y = (img_y - y0) / (y1 - y0) * 2 - 1
            img_x = (img_x - x0) / (x1 - x0) * 2 - 1
            if torch.isinf(img_x).any():
                inds = torch.where(torch.isinf(img_x))
                img_x[inds] = 0
            if torch.isinf(img_y).any():
                inds = torch.where(torch.isinf(img_y))
                img_y[inds] = 0

            gx = img_x[None, :].expand(img_y.size(0), img_x.size(0))
            gy = img_y[:, None].expand(img_y.size(0), img_x.size(0))
            grid = torch.stack([gx, gy], dim=2)

            img_masks = F.grid_sample(
                mask.to(dtype=torch.float32)[None, None, :, :],
                grid[None, :, :, :],
                align_corners=False,
            )

            result_masks.append(img_masks)
        result_masks = torch.cat(result_masks, 1)
        return result_masks.squeeze(0)

    def forward(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
        mode: str = "predict",
        **kwargs,
    ) -> Any:
        """The model forward.

        Args:
            inputs (torch.Tensor): The input tensors
            data_samples (List[BaseDataElement], optional): The data samples.
                Defaults to None.
            mode (str, optional): forward mode, only support `predict`.

        Returns:
            Any: Model output.
        """
        assert mode == "predict", 'Deploy model only allow mode=="predict".'
        inputs = inputs.contiguous()
        outputs = self.predict(inputs)
        outputs = End2EndModel.__clear_outputs(outputs)
        batch_dets, batch_labels = outputs[:2]
        batch_masks = outputs[2] if len(outputs) == 3 else None
        batch_size = inputs.shape[0]
        img_metas = [data_sample.metainfo for data_sample in data_samples]
        results = []
        rescale = kwargs.get("rescale", True)
        model_type = self.model_cfg.model.type if self.model_cfg is not None else None
        for i in range(batch_size):
            dets, labels = batch_dets[i], batch_labels[i]
            result = InstanceData()

            bboxes = dets[:, :4]
            scores = dets[:, 4]
            # perform rescale
            if rescale and "scale_factor" in img_metas[i]:
                scale_factor = img_metas[i]["scale_factor"]
                if isinstance(scale_factor, (list, tuple, np.ndarray)):
                    if len(scale_factor) == 2:
                        scale_factor = np.array(scale_factor)
                        scale_factor = np.concatenate([scale_factor, scale_factor])
                    scale_factor = np.array(scale_factor)[None, :]  # [1,4]
                scale_factor = torch.from_numpy(scale_factor).to(dets)
                bboxes /= scale_factor

            # Most of models in mmdetection 3.x use `pad_param`, but some
            # models like CenterNet uses `border`.
            # offset pixel of the top-left corners between original image
            # and padded/enlarged image, 'pad_param' is used when exporting
            # CornerNet and CentripetalNet to onnx
            pad_key = None
            if "pad_param" in img_metas[i]:
                pad_key = "pad_param"
            elif "border" in img_metas[i]:
                pad_key = "border"
            if pad_key is not None:
                scale_factor = img_metas[i].get("scale_factor", np.array([1.0, 1.0]))
                x_off = img_metas[i][pad_key][2] / scale_factor[1]
                y_off = img_metas[i][pad_key][0] / scale_factor[0]
                bboxes[:, ::2] -= x_off
                bboxes[:, 1::2] -= y_off
                bboxes *= bboxes > 0

            result.scores = scores
            result.bboxes = bboxes
            if model_type in ["SOLO", "SOLOv2"]:
                result.bboxes = bboxes.new_zeros(bboxes.shape)
            result.labels = labels

            if batch_masks is not None:
                masks = batch_masks[i]
                img_h, img_w = img_metas[i]["img_shape"][:2]
                ori_h, ori_w = img_metas[i]["ori_shape"][:2]
                export_postprocess_mask = False
                if self.deploy_cfg is not None:
                    mmdet_deploy_cfg = get_post_processing_params(self.deploy_cfg)
                    # this flag enable postprocess when export.
                    export_postprocess_mask = mmdet_deploy_cfg.get(
                        "export_postprocess_mask", False
                    )
                if not export_postprocess_mask:
                    masks = End2EndModel.postprocessing_masks(
                        dets[:, :4], masks, ori_w, ori_h, self.device
                    )
                else:
                    masks = masks[:, :img_h, :img_w]
                # avoid to resize masks with zero dim
                if export_postprocess_mask and rescale and masks.shape[0] != 0:
                    masks = torch.nn.functional.interpolate(
                        masks.unsqueeze(0),
                        size=[
                            math.ceil(
                                masks.shape[-2] / img_metas[i]["scale_factor"][0]
                            ),
                            math.ceil(
                                masks.shape[-1] / img_metas[i]["scale_factor"][1]
                            ),
                        ],
                    )[..., :ori_h, :ori_w]
                    masks = masks.squeeze(0)
                if masks.dtype != bool:
                    masks = masks >= 0.5
                # aligned with mmdet to easily convert to numpy
                masks = masks.cpu()
                result.masks = masks
            data_samples[i].pred_instances = result
            results.append(data_samples[i])
        return results

    def predict(self, imgs: Tensor) -> Tuple[np.ndarray, np.ndarray]:
        """The interface for predict.

        Args:
            imgs (Tensor): Input image(s) in [N x C x H x W] format.

        Returns:
            tuple[np.ndarray, np.ndarray]: dets of shape [N, num_det, 5]
                and class labels of shape [N, num_det].
        """
        outputs = self.wrapper({self.input_name: imgs})
        outputs = self.wrapper.output_to_list(outputs)
        return outputs


@__BACKEND_MODEL.register_module("single_stage")
class PartitionSingleStageModel(End2EndModel):
    """Partitioned single stage detection model.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string specifying device type.
        model_cfg (str|Config): Input model config file or Config
            object.
        deploy_cfg (str|Config): Deployment config file or loaded Config
            object.
        data_preprocessor (dict|nn.Module): The data preprocessor.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        model_cfg: Union[str, Config],
        deploy_cfg: Union[str, Config],
        data_preprocessor: Optional[Union[dict, nn.Module]] = None,
        **kwargs,
    ):
        super().__init__(
            backend, backend_files, device, deploy_cfg, data_preprocessor, **kwargs
        )
        # load cfg if necessary
        model_cfg = load_config(model_cfg)[0]
        self.model_cfg = model_cfg

    def _init_wrapper(self, backend, backend_files, device):
        self.wrapper = BaseBackendModel._build_wrapper(
            backend=backend,
            backend_files=backend_files,
            device=device,
            output_names=["scores", "boxes"],
            deploy_cfg=self.deploy_cfg,
        )

    def partition0_postprocess(self, scores: Tensor, bboxes: Tensor):
        """Perform post-processing for partition 0.

        Args:
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            bboxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].

        Returns:
            tuple[np.ndarray, np.ndarray]: dets of shape [N, num_det, 5] and
                class labels of shape [N, num_det].
        """
        cfg = self.model_cfg.model.test_cfg
        deploy_cfg = self.deploy_cfg

        post_params = get_post_processing_params(deploy_cfg)
        max_output_boxes_per_class = post_params.max_output_boxes_per_class
        iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
        score_threshold = cfg.get("score_thr", post_params.score_threshold)
        pre_top_k = (
            -1 if post_params.pre_top_k >= bboxes.shape[1] else post_params.pre_top_k
        )
        keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
        ret = multiclass_nms(
            bboxes,
            scores,
            max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
        )
        ret = [r.cpu() for r in ret]
        return ret

    def predict(self, imgs: Tensor, *args, **kwargs):
        """Implement forward test.

        Args:
            imgs (Tensor): Input image(s) in [N x C x H x W] format.

        Returns:
            list[np.ndarray, np.ndarray]: dets of shape [N, num_det, 5] and
                class labels of shape [N, num_det].
        """
        outputs = self.wrapper({self.input_name: imgs})
        outputs = self.wrapper.output_to_list(outputs)
        scores, bboxes = outputs[:2]
        return self.partition0_postprocess(scores, bboxes)


@__BACKEND_MODEL.register_module("two_stage")
class PartitionTwoStageModel(End2EndModel):
    """Partitioned two stage detection model.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string specifying device type.
        model_cfg (str|Config): Input model config file or Config
            object.
        deploy_cfg (str|Config): Deployment config file or loaded Config
            object.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        model_cfg: Union[str, Config],
        deploy_cfg: Union[str, Config],
        data_preprocessor: Optional[Union[dict, nn.Module]] = None,
        **kwargs,
    ):
        # load cfg if necessary
        model_cfg = load_config(model_cfg)[0]

        self.model_cfg = model_cfg

        super().__init__(
            backend, backend_files, device, deploy_cfg, data_preprocessor, **kwargs
        )
        from mmdet.models.builder import build_head, build_roi_extractor

        from ..models.roi_heads.bbox_head import bbox_head__get_bboxes

        self.bbox_roi_extractor = build_roi_extractor(
            model_cfg.model.roi_head.bbox_roi_extractor
        )
        self.bbox_head = build_head(model_cfg.model.roi_head.bbox_head)

        class Context:
            pass

        ctx = Context()
        ctx.cfg = self.deploy_cfg
        self.bbox_head__get_bboxes = partial(bbox_head__get_bboxes, ctx)

    def _init_wrapper(self, backend, backend_files, device):
        n = get_backend_file_count(backend)
        num_feat = self.model_cfg["model"]["neck"]["num_outs"]
        partition0_output_names = ["feat/{}".format(i) for i in range(num_feat)] + [
            "scores",
            "boxes",
        ]

        self.first_wrapper = BaseBackendModel._build_wrapper(
            backend,
            backend_files[0:n],
            device,
            output_names=partition0_output_names,
            deploy_cfg=self.deploy_cfg,
        )

        self.second_wrapper = BaseBackendModel._build_wrapper(
            backend,
            backend_files[n : 2 * n],
            device,
            output_names=["cls_score", "bbox_pred"],
            deploy_cfg=self.deploy_cfg,
        )

    def partition0_postprocess(
        self, x: Sequence[Tensor], scores: Tensor, bboxes: Tensor
    ):
        """Perform post-processing for partition 0.

        Args:
            x (tuple[Tensor]): Feature maps of all scale levels.
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            bboxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].

        Returns:
            tuple(Tensor, Tensor): rois and bbox_feats.
        """
        # rpn-nms + roi-extractor
        cfg = self.model_cfg.model.test_cfg.rpn
        deploy_cfg = self.deploy_cfg

        post_params = get_post_processing_params(deploy_cfg)
        iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
        score_threshold = cfg.get("score_thr", post_params.score_threshold)
        pre_top_k = (
            -1 if post_params.pre_top_k >= bboxes.shape[1] else post_params.pre_top_k
        )
        keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
        # only one class in rpn
        max_output_boxes_per_class = keep_top_k
        proposals, _ = multiclass_nms(
            bboxes,
            scores,
            max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
        )

        rois = proposals
        batch_index = (
            torch.arange(rois.shape[0], device=rois.device)
            .float()
            .view(-1, 1, 1)
            .expand(rois.size(0), rois.size(1), 1)
        )
        rois = torch.cat([batch_index, rois[..., :4]], dim=-1)
        batch_size = rois.shape[0]
        num_proposals_per_img = rois.shape[1]

        # Eliminate the batch dimension
        rois = rois.view(-1, 5)
        bbox_feats = self.bbox_roi_extractor(
            x[: self.bbox_roi_extractor.num_inputs], rois
        )

        rois = rois.reshape(batch_size, num_proposals_per_img, rois.size(-1))
        return rois, bbox_feats

    def partition1_postprocess(
        self,
        rois: Tensor,
        cls_score: Tensor,
        bbox_pred: Tensor,
        img_metas: Sequence[dict],
    ):
        """Perform post-processing for partition 1.
        Args:
            rois (Tensor): Input tensor of roi.
            cls_score (Tensor): Scores of all classes.
            bbox_pred (Tensor): Bounding box proposals.
            img_metas (Sequence[dict]): A list of image(s) meta information.

        Returns:
            tuple[Tensor, Tensor]: dets of shape [N, num_det, 5] and class
                labels of shape [N, num_det].
        """
        batch_size = rois.shape[0]
        num_proposals_per_img = rois.shape[1]

        cls_score = cls_score.reshape(
            batch_size, num_proposals_per_img, cls_score.size(-1)
        )

        bbox_pred = bbox_pred.reshape(
            batch_size, num_proposals_per_img, bbox_pred.size(-1)
        )

        rcnn_test_cfg = self.model_cfg.model.test_cfg.rcnn
        return self.bbox_head__get_bboxes(
            self.bbox_head,
            rois,
            cls_score,
            bbox_pred,
            img_metas[0][0]["img_shape"],
            img_metas[0][0]["scale_factor"],
            cfg=rcnn_test_cfg,
        )

    def predict(self, imgs: Tensor, img_metas: Sequence[dict], *args, **kwargs):
        """Implement forward test.

        Args:
            imgs (Tensor): Input image(s) in [N x C x H x W] format.
            img_metas (Sequence[dict]): A list of image(s) meta information.

        Returns:
            tuple[np.ndarray, np.ndarray]: dets of shape [N, num_det, 5] and
                class labels of shape [N, num_det].
        """
        outputs = self.first_wrapper({"input": imgs})
        outputs = self.first_wrapper.output_to_list(outputs)
        feats = outputs[:-2]
        scores, bboxes = outputs[-2:]

        # partition0_postprocess
        rois, bbox_feats = self.partition0_postprocess(feats, scores, bboxes)

        # partition1 forward
        bbox_feats = bbox_feats.contiguous()
        outputs = self.second_wrapper({"bbox_feats": bbox_feats})
        outputs = self.second_wrapper.output_to_list(outputs)
        cls_score, bbox_pred = outputs[:2]

        # partition1_postprocess
        outputs = self.partition1_postprocess(rois, cls_score, bbox_pred, img_metas)
        outputs = [out.detach().cpu() for out in outputs]
        return outputs


@__BACKEND_MODEL.register_module("ncnn_end2end")
class NCNNEnd2EndModel(End2EndModel):
    """NCNNEnd2EndModel.

    End2end NCNN model inference class. Because it has DetectionOutput layer
    and its output is different from original mmdet style of `dets`, `labels`.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string specifying device type.
        model_cfg (str|Config): Input model config file or Config
            object.
        deploy_cfg (str|Config): Deployment config file or loaded Config
            object.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        model_cfg: Union[str, Config],
        deploy_cfg: Union[str, Config],
        **kwargs,
    ):
        assert (
            backend == Backend.NCNN
        ), f"only supported ncnn, but give \
            {backend.value}"

        super(NCNNEnd2EndModel, self).__init__(
            backend, backend_files, device, deploy_cfg, **kwargs
        )
        # load cfg if necessary
        model_cfg = load_config(model_cfg)[0]
        self.model_cfg = model_cfg

    def predict(self, imgs: Tensor, *args, **kwargs) -> List:
        """Implement forward test.

        Args:
            imgs (Tensor): Input image(s) in [N x C x H x W] format.

        Returns:
            list[Tensor]: dets of shape [N, num_det, 5] and
                class labels of shape [N, num_det].
        """
        _, _, H, W = imgs.shape
        outputs = self.wrapper({self.input_name: imgs})
        for key, item in outputs.items():
            if item is None:
                return torch.zeros(1, 0, 5), torch.zeros(1, 0)
        out = self.wrapper.output_to_list(outputs)[0]
        labels = out[:, :, 0] - 1
        scales = torch.tensor([W, H, W, H]).reshape(1, 1, 4).to(out)
        scores = out[:, :, 1:2]
        boxes = out[:, :, 2:6] * scales
        dets = torch.cat([boxes, scores], dim=2)
        return dets, labels.to(torch.int32)


@__BACKEND_MODEL.register_module("vacc_det")
class VACCDetModel(End2EndModel):
    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        model_cfg: Union[str, Config],
        deploy_cfg: Union[str, Config],
        **kwargs,
    ):
        assert backend == Backend.VACC, (
            "only supported vacc, but give " f"{backend.value}"
        )

        super(VACCDetModel, self).__init__(
            backend, backend_files, device, deploy_cfg, **kwargs
        )
        # load cfg if necessary
        model_cfg = load_config(model_cfg)[0]
        self.model_cfg = model_cfg

    def forward(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
        mode: str = "predict",
        **kwargs,
    ) -> Any:
        """The model forward.

        Args:
            inputs (torch.Tensor): The input tensors
            data_samples (List[BaseDataElement], optional): The data samples.
                Defaults to None.
            mode (str, optional): forward mode, only support `predict`.
        Returns:
            Any: Model output.
        """
        assert mode == "predict", 'Deploy model only allow mode=="predict".'
        inputs = inputs.contiguous()
        outputs = self.wrapper({self.input_name: inputs})
        outputs = [i for i in outputs.values()]
        ret = self._get_bboxes(outputs[0], [i.metainfo for i in data_samples])
        for i in range(len(ret)):
            data_samples[i].pred_instances = ret[i]
        return data_samples

    def _get_bboxes(self, outputs, img_metas):
        from mmdet.registry import MODELS

        head_cfg = self.model_cfg._cfg_dict.model.bbox_head
        head = MODELS.build(head_cfg)
        if head_cfg.type == "YOLOV3Head":
            ret = head.predict_by_feat(
                outputs,
                [dict(scale_factor=None)],
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
            )
        else:
            raise NotImplementedError(f"{head_cfg.type} not supported yet.")
        return ret


@__BACKEND_MODEL.register_module("sdk")
class SDKEnd2EndModel(End2EndModel):
    """SDK inference class, converts SDK output to mmdet format."""

    def __init__(self, *args, **kwargs):
        kwargs["data_preprocessor"] = None
        super(SDKEnd2EndModel, self).__init__(*args, **kwargs)
        self.has_mask = self.deploy_cfg.codebase_config.get("has_mask", False)

    def forward(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
        mode: str = "predict",
    ):
        """Run forward inference.

        Args:
            inputs (Sequence[Tensor]): A list contains input image(s)
                in [C x H x W] format.
            data_samples (List[BaseDataElement]): A list of meta info
                for image(s).
            *args: Other arguments.
            **kwargs: Other key-pair arguments.

        Returns:
            list: A list contains predictions.
        """
        if isinstance(inputs, list):
            inputs = inputs[0]
        # inputs are c,h,w, sdk requested h,w,c
        inputs = inputs.permute(1, 2, 0)
        dets, labels, masks = self.wrapper.invoke(
            inputs.contiguous().detach().cpu().numpy()
        )
        result = InstanceData()
        if self.has_mask:
            segm_results = []
            ori_h, ori_w = data_samples[0].ori_shape[:2]
            for bbox, mask in zip(dets, masks):
                img_mask = np.zeros((ori_h, ori_w), dtype=np.uint8)
                left = int(max(np.floor(bbox[0]) - 1, 0))
                top = int(max(np.floor(bbox[1]) - 1, 0))
                img_mask[top : top + mask.shape[0], left : left + mask.shape[1]] = mask
                segm_results.append(torch.from_numpy(img_mask))
            if len(segm_results) > 0:
                result.masks = torch.stack(segm_results, 0).to(self.device)
            else:
                result.masks = torch.zeros([0, ori_h, ori_w]).to(self.device)
        dets = torch.from_numpy(dets).to(self.device)
        labels = torch.from_numpy(labels).to(torch.int64).to(self.device)
        result.bboxes = dets[:, :4]
        result.scores = dets[:, 4]
        result.labels = labels
        data_samples[0].pred_instances = result

        return data_samples


@__BACKEND_MODEL.register_module("rknn")
class RKNNModel(End2EndModel):
    """RKNNModel.

    RKNN inference class, converts RKNN output to mmdet format.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string specifying device type.
        model_cfg (str|Config): Input model config file or Config
            object.
        deploy_cfg (str|Config): Deployment config file or loaded Config
            object.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        model_cfg: Union[str, Config],
        deploy_cfg: Union[str, Config],
        **kwargs,
    ):
        assert (
            backend == Backend.RKNN
        ), f"only supported rknn, but give \
            {backend.value}"

        super(RKNNModel, self).__init__(
            backend, backend_files, device, deploy_cfg, **kwargs
        )
        # load cfg if necessary
        model_cfg = load_config(model_cfg)[0]
        self.model_cfg = model_cfg
        self.batch_size = get_quantization_config(deploy_cfg).get("rknn_batch_size", -1)

    def _init_wrapper(self, backend, backend_files, device):
        """Initialize backend wrapper.

        Args:
            backend (Backend): The backend enum, specifying backend type.
            backend_files (Sequence[str]): Paths rknn model files.
            device (str): A string specifying device type.
        """
        output_names = None
        if self.deploy_cfg is not None:
            ir_config = get_ir_config(self.deploy_cfg)
            output_names = ir_config.get("output_names", None)
            if get_partition_config(self.deploy_cfg) is not None:
                output_names = get_partition_config(self.deploy_cfg)["partition_cfg"][
                    0
                ]["output_names"]

        self.wrapper = BaseBackendModel._build_wrapper(
            backend,
            backend_files,
            device,
            output_names=output_names,
            deploy_cfg=self.deploy_cfg,
        )

    def _get_bboxes(self, outputs: List[Tensor], metainfos: Any):
        """get bboxes from output by meta infos.

        Args:
            outputs (List[Tensor]): The backend wrapper outputs.
            metainfos (Any): The meta infos of inputs.

        Raises:
            NotImplementedError: Head type not supported.

        Returns:
            Any: model outputs.
        """
        from mmdet.models import build_head

        head_cfg = self.model_cfg._cfg_dict.model.bbox_head
        head = build_head(head_cfg)
        if head_cfg.type == "YOLOXHead":
            divisor = round(len(outputs) / 3)
            ret = head.predict_by_feat(
                outputs[:divisor],
                outputs[divisor : 2 * divisor],
                outputs[2 * divisor :],
                metainfos,
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
                rescale=True,
            )
        elif head_cfg.type == "YOLOV3Head":
            ret = head.predict_by_feat(
                outputs,
                metainfos,
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
                rescale=True,
            )
        elif head_cfg.type == "YOLOv5Head":
            from mmdet.models.utils import multi_apply

            outs = multi_apply(
                self.yolov5_head_forward_single,
                outputs,
                [head.num_base_priors] * len(outputs),
            )
            ret = head.predict_by_feat(
                *outs,
                batch_img_metas=metainfos,
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
                rescale=True,
            )
        elif head_cfg.type == "RTMDetSepBNHead":
            divisor = round(len(outputs) / 2)
            ret = head.predict_by_feat(
                outputs[:divisor],
                outputs[divisor:],
                batch_img_metas=metainfos,
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
                rescale=True,
            )
        elif head_cfg.type in ("RetinaHead", "SSDHead", "FSAFHead"):
            partition_cfgs = get_partition_config(self.deploy_cfg)
            if partition_cfgs is None:  # bbox decoding done in rknn model
                from mmdet.structures.bbox import scale_boxes

                from ..models.layers.bbox_nms import _multiclass_nms

                dets, labels = _multiclass_nms(outputs[0], outputs[1])
                ret = [InstanceData() for i in range(dets.shape[0])]
                for i, instance_data in enumerate(ret):
                    instance_data.bboxes = dets[i, :, :4]
                    instance_data.scores = dets[i, :, 4]
                    instance_data.labels = labels[i]
                    scale_factor = [1 / s for s in metainfos[i]["scale_factor"]]
                    instance_data.bboxes = scale_boxes(
                        instance_data.bboxes, scale_factor
                    )
                return ret
            divisor = round(len(outputs) / 2)
            ret = head.predict_by_feat(
                outputs[:divisor],
                outputs[divisor:],
                batch_img_metas=metainfos,
                rescale=True,
                cfg=self.model_cfg._cfg_dict.model.test_cfg,
            )
        else:
            raise NotImplementedError(f"{head_cfg.type} not supported yet.")
        return ret

    def forward(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
        mode: str = "predict",
        *args,
        **kwargs,
    ):
        """Implement forward test.

        Args:
            inputs (Tensor): Inputs with shape (N, C, H, W).
            data_samples (List[:obj:`DetDataSample`]): The Data
                Samples. It usually includes information such as
                `gt_instance`, `gt_panoptic_seg` and `gt_sem_seg`.

        Returns:
            list[np.ndarray, np.ndarray]: dets of shape [N, num_det, 5] and
                class labels of shape [N, num_det].
        """
        if self.batch_size > len(data_samples):
            pad_len = self.batch_size - len(data_samples)
            data_samples = data_samples + [data_samples[0]] * pad_len
            inputs = torch.cat([inputs] + [inputs[:1, ...]] * pad_len, 0)
        outputs = self.wrapper({self.input_name: inputs})
        outputs = [i for i in outputs.values()]
        ret = self._get_bboxes(outputs, [i.metainfo for i in data_samples])
        for i in range(len(ret)):
            data_samples[i].pred_instances = ret[i]
        return data_samples

    def yolov5_head_forward_single(
        self, pred_map: Tensor, num_base_priors: int
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward feature of a single scale level."""
        bs, _, ny, nx = pred_map.shape
        pred_map = pred_map.view(bs, num_base_priors, -1, ny, nx)

        cls_score = pred_map[:, :, 5:, ...].reshape(bs, -1, ny, nx)
        bbox_pred = pred_map[:, :, :4, ...].reshape(bs, -1, ny, nx)
        objectness = pred_map[:, :, 4:5, ...].reshape(bs, -1, ny, nx)

        return cls_score, bbox_pred, objectness


def build_object_detection_model(
    model_files: Sequence[str],
    model_cfg: Union[str, Config],
    deploy_cfg: Union[str, Config],
    device: str,
    data_preprocessor: Optional[Union[Config, BaseDataPreprocessor]] = None,
    **kwargs,
):
    """Build object detection model for different backends.

    Args:
        model_files (Sequence[str]): Input model file(s).
        model_cfg (str | Config): Input model config file or Config
            object.
        deploy_cfg (str | Config): Input deployment config file or
            Config object.
        device (str):  Device to input model
        data_preprocessor (BaseDataPreprocessor | Config): The data
            preprocessor of the model.

    Returns:
        End2EndModel: Detector for a configured backend.
    """
    # load cfg if necessary
    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

    backend = get_backend(deploy_cfg)

    partition_config = get_partition_config(deploy_cfg)
    if partition_config is not None:
        partition_type = partition_config.get("type", None)
    else:
        codebase_config = get_codebase_config(deploy_cfg)
        # Default Config is 'end2end'
        partition_type = codebase_config.get("model_type", "end2end")

    backend_detector = __BACKEND_MODEL.build(
        dict(
            type=partition_type,
            backend=backend,
            backend_files=model_files,
            device=device,
            model_cfg=model_cfg,
            deploy_cfg=deploy_cfg,
            data_preprocessor=data_preprocessor,
            **kwargs,
        )
    )

    return backend_detector



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
from typing import Dict, List, Optional, Union

import mmengine
from torch import nn

from juxtapose.mmdeploy.utils.constants import IR, Backend
from .rewriter_utils import Checker, RewriterRegistry, collect_env, eval_with_import


class ModuleRewriter:
    """A module rewriter which maintains rewritten modules.

    The rewritten modules can be registered by calling
    register_rewrite_module().  By calling patch_model(), all the registered
    modules of model will be replaced.

    Examples:
        >>> @MODULE_REWRITER.register_rewrite_module(
        >>>     'mmagic.models.backbones.sr_backbones.SRCNN',
        >>>     backend='tensorrt')
        >>> class SRCNNWrapper(torch.nn.Module):
        >>>     # rewrite the module here
    """

    def __init__(self):
        self._registry = RewriterRegistry()

    def register_rewrite_module(
        self,
        module_type: str,
        backend: str = Backend.DEFAULT.value,
        ir: IR = IR.DEFAULT,
        extra_checkers: Optional[Union[Checker, List[Checker]]] = None,
        **kwargs,
    ):
        """The interface of module rewriter decorator.

        Args:
            module_type (str): The module type name to rewrite.
            backend (str): The rewriter will be activated on which backend.
            ir (IR): The rewriter will be activated on which IR.
            extra_checkers (Checker | List[Checker] | None): Other requirements
                defined by Checker.

        Returns:
            nn.Module: The rewritten model.
        """
        return self._registry.register_object(
            module_type, backend, ir, extra_checkers, **kwargs
        )

    def patch_model(
        self,
        model: nn.Module,
        cfg: mmengine.Config,
        backend: str = Backend.DEFAULT.value,
        ir: IR = IR.DEFAULT,
        recursive: bool = True,
        **kwargs,
    ) -> nn.Module:
        """Replace the models that was registered.

        Args:
            model (torch.nn.Module): The model to patch.
            cfg (Dict): Config dictionary of deployment.
            backend (str): The inference engine name.
            ir (IR): The intermeditate representation name.
            recursive (bool): The flag to enable recursive patching.

        Returns:
            nn.Module: THe patched model.

        Examples:
            >>> from juxtapose.mmdeploy.core import patch_model
            >>> patched_model = patch_model(model, cfg=deploy_cfg,
            >>>                             backend=backend)
        """
        # TODO: Make the type of parameter backend to Backend
        env = collect_env(Backend.get(backend), ir)
        self._collect_record(env)
        return self._replace_module(model, cfg, recursive, **kwargs)

    def _replace_one_module(self, module, cfg, **kwargs):
        """Build a rewritten model."""
        # module could be instance of multiple classes
        object_dict_candidate = dict()
        for k, v in self._records.items():
            if isinstance(module, k):
                object_dict_candidate[k] = v
        if len(object_dict_candidate) == 0:
            return module

        type_sequence = [type(module)]
        while len(type_sequence) > 0:
            # pop if type is object
            module_type = type_sequence.pop(0)
            if module_type == object:
                continue
            object_dict = object_dict_candidate.get(module_type, None)
            if object_dict is not None:
                break
            else:
                type_sequence.extend(module_type.__bases__)

        module_class = object_dict["_object"]

        # Pop arguments that are not supported
        input_args = kwargs.copy()
        supported_args = inspect.getfullargspec(module_class.__init__).args
        redundant_key_name = []
        for k in input_args:
            if k not in supported_args:
                redundant_key_name.append(k)
        for k in redundant_key_name:
            input_args.pop(k)

        return module_class(module, cfg, **input_args)

    def _replace_module(
        self, model: nn.Module, cfg: mmengine.Config, recursive: bool, **kwargs
    ):
        """DFS and replace target models."""

        def _replace_module_impl(model, cfg, **kwargs):
            # disable patching if model is already patched.
            if type(model) in self._cls_set:
                return model

            if recursive and hasattr(model, "named_children"):
                for name, module in model.named_children():
                    model._modules[name] = _replace_module_impl(module, cfg, **kwargs)
            return self._replace_one_module(model, cfg, **kwargs)

        return _replace_module_impl(model, cfg, **kwargs)

    def _collect_record(self, env: Dict):
        """Collect models in registry."""
        self._records = {}
        self._cls_set = set()
        records = self._registry.get_records(env)
        for name, kwargs in records:
            self._cls_set.add(kwargs["_object"])
            self._records[eval_with_import(name)] = kwargs



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmengine.config import ConfigDict
from mmengine.structures import InstanceData
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.yolox_head." "YOLOXHead.predict_by_feat"
)
def yolox_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    objectnesses: Optional[List[Tensor]],
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
) -> List[InstanceData]:
    """Rewrite `predict_by_feat` of `YOLOXHead` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx: Context that contains original meta information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        objectnesses (list[Tensor], Optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, 1, H, W).
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.

    Returns:
        tuple[Tensor, Tensor]: The first item is an (N, num_box, 5) tensor,
            where 5 represent (tl_x, tl_y, br_x, br_y, score), N is batch
            size and the score between 0 and 1. The shape of the second
            tensor in the tuple is (N, num_box), and each element
            represents the class label of the corresponding box.
    """
    ctx = FUNCTION_REWRITER.get_context()

    # mark pred_maps
    @mark("yolo_head", inputs=["cls_scores", "bbox_preds", "objectnesses"])
    def __mark_pred_maps(cls_scores, bbox_preds, objectnesses):
        return cls_scores, bbox_preds, objectnesses

    cls_scores, bbox_preds, objectnesses = __mark_pred_maps(
        cls_scores, bbox_preds, objectnesses
    )
    assert len(cls_scores) == len(bbox_preds) == len(objectnesses)
    device = cls_scores[0].device
    cfg = self.test_cfg if cfg is None else cfg
    batch_size = bbox_preds[0].shape[0]
    featmap_sizes = [cls_score.shape[2:] for cls_score in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, device=device, with_stride=True
    )

    flatten_cls_scores = [
        cls_score.permute(0, 2, 3, 1).reshape(batch_size, -1, self.cls_out_channels)
        for cls_score in cls_scores
    ]
    flatten_bbox_preds = [
        bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4)
        for bbox_pred in bbox_preds
    ]
    flatten_objectness = [
        objectness.permute(0, 2, 3, 1).reshape(batch_size, -1)
        for objectness in objectnesses
    ]

    cls_scores = torch.cat(flatten_cls_scores, dim=1).sigmoid()
    score_factor = torch.cat(flatten_objectness, dim=1).sigmoid()
    flatten_bbox_preds = torch.cat(flatten_bbox_preds, dim=1)
    flatten_priors = torch.cat(mlvl_priors)
    bboxes = self._bbox_decode(flatten_priors, flatten_bbox_preds)
    # directly multiply score factor and feed to nms
    scores = cls_scores * (score_factor.unsqueeze(-1))

    if not with_nms:
        return bboxes, scores

    deploy_cfg = ctx.cfg
    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)

    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        bboxes,
        scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.yolox_head." "YOLOXHead.predict_by_feat",
    backend=Backend.NCNN.value,
)
def yolox_head__predict_by_feat__ncnn(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    objectnesses: Optional[List[Tensor]],
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
):
    """Rewrite `predict_by_feat` of YOLOXHead for ncnn backend.

    1. Decode the prior to a box format for ncnn DetectionOutput layer to do
    the post-processing.
    2. Batch dimension is not supported by ncnn, but supported by pytorch.
    The negative value of axis in torch.cat is rewritten as corresponding
    positive value to avoid axis shift.
    3. 2-dimension tensor broadcast of `BinaryOps` operator is not supported by
    ncnn. This function unsqueeze 2-dimension tensor to 3-dimension tensor for
    correct `BinaryOps` calculation by ncnn.

    Args:
        ctx: Context that contains original meta information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        objectnesses (list[Tensor], Optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, 1, H, W).
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.

    Returns:
        output__ncnn (Tensor): outputs, shape is [N, num_det, 6].
    """
    ctx = FUNCTION_REWRITER.get_context()
    from juxtapose.mmdeploy.codebase.mmdet.ops import ncnn_detection_output_forward
    from juxtapose.mmdeploy.utils import get_root_logger
    from juxtapose.mmdeploy.utils.config_utils import is_dynamic_shape

    dynamic_flag = is_dynamic_shape(ctx.cfg)
    if dynamic_flag:
        logger = get_root_logger()
        logger.warning("YOLOX does not support dynamic shape with ncnn.")
    img_height = int(batch_img_metas[0]["img_shape"][0])
    img_width = int(batch_img_metas[0]["img_shape"][1])

    assert len(cls_scores) == len(bbox_preds) == len(objectnesses)
    device = cls_scores[0].device
    cfg = self.test_cfg if cfg is None else cfg
    batch_size = bbox_preds[0].shape[0]
    featmap_sizes = [cls_score.shape[2:] for cls_score in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, device=device, with_stride=True
    )
    mlvl_priors = [mlvl_prior.unsqueeze(0) for mlvl_prior in mlvl_priors]
    flatten_priors = torch.cat(mlvl_priors, dim=1)

    flatten_cls_scores = [
        cls_score.permute(0, 2, 3, 1).reshape(batch_size, -1, self.cls_out_channels)
        for cls_score in cls_scores
    ]
    flatten_bbox_preds = [
        bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4)
        for bbox_pred in bbox_preds
    ]
    flatten_objectness = [
        objectness.permute(0, 2, 3, 1).reshape(batch_size, -1, 1)
        for objectness in objectnesses
    ]

    cls_scores = torch.cat(flatten_cls_scores, dim=1).sigmoid()
    dummy_cls_scores = torch.zeros(
        batch_size, cls_scores.shape[-2], 1, device=cls_scores.device
    )

    batch_mlvl_scores = torch.cat([dummy_cls_scores, cls_scores], dim=2)
    score_factor = torch.cat(flatten_objectness, dim=1).sigmoid()
    flatten_bbox_preds = torch.cat(flatten_bbox_preds, dim=1)
    assert (
        flatten_priors.shape[-1] == 4
    ), f"yolox needs (B, N, 4) priors, got\
        (B, N, {flatten_priors.shape[-1]})"
    prior_box_x1 = (
        flatten_priors[:, :, 0:1] - flatten_priors[:, :, 2:3] / 2
    ) / img_width
    prior_box_y1 = (
        flatten_priors[:, :, 1:2] - flatten_priors[:, :, 3:4] / 2
    ) / img_height
    prior_box_x2 = (
        flatten_priors[:, :, 0:1] + flatten_priors[:, :, 2:3] / 2
    ) / img_width
    prior_box_y2 = (
        flatten_priors[:, :, 1:2] + flatten_priors[:, :, 3:4] / 2
    ) / img_height
    prior_box_ncnn = torch.cat(
        [prior_box_x1, prior_box_y1, prior_box_x2, prior_box_y2], dim=2
    )

    scores = batch_mlvl_scores.permute(0, 2, 1).unsqueeze(3) * score_factor.permute(
        0, 2, 1
    ).unsqueeze(3)
    scores = scores.squeeze(3).permute(0, 2, 1)

    batch_mlvl_bboxes = flatten_bbox_preds.reshape(batch_size, 1, -1)
    batch_mlvl_scores = scores.reshape(batch_size, 1, -1)
    batch_mlvl_priors = prior_box_ncnn.reshape(batch_size, 1, -1)
    batch_mlvl_vars = torch.ones_like(batch_mlvl_priors)
    batch_mlvl_priors = torch.cat([batch_mlvl_priors, batch_mlvl_vars], dim=1)
    deploy_cfg = ctx.cfg
    post_params = get_post_processing_params(deploy_cfg)
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)

    vars = torch.tensor([1, 1, 1, 1], dtype=torch.float32)
    output__ncnn = ncnn_detection_output_forward(
        batch_mlvl_bboxes,
        batch_mlvl_scores,
        batch_mlvl_priors,
        score_threshold,
        iou_threshold,
        pre_top_k,
        keep_top_k,
        self.num_classes + 1,
        vars.cpu().detach().numpy(),
    )
    return output__ncnn



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
from typing import Any, Callable, Dict, Optional, Sequence

import torch

from juxtapose.mmdeploy.core.rewriters import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import IR, cfg_apply_marks, get_partition_config

MARK_FUNCTION_COUNT = dict()


def reset_mark_function_count():
    """Reset counter of mark function."""
    for k in MARK_FUNCTION_COUNT:
        MARK_FUNCTION_COUNT[k] = 0


TORCH_DTYPE_TO_ONNX = {
    torch.uint8: torch.onnx.TensorProtoDataType.UINT8,
    torch.int8: torch.onnx.TensorProtoDataType.INT8,
    torch.float64: torch.onnx.TensorProtoDataType.DOUBLE,
    torch.float32: torch.onnx.TensorProtoDataType.FLOAT,
    torch.float16: torch.onnx.TensorProtoDataType.FLOAT16,
    torch.int32: torch.onnx.TensorProtoDataType.INT32,
    torch.int64: torch.onnx.TensorProtoDataType.INT64,
    torch.int16: torch.onnx.TensorProtoDataType.INT16,
    torch.bool: torch.onnx.TensorProtoDataType.BOOL,
    torch.complex64: torch.onnx.TensorProtoDataType.COMPLEX64,
    torch.complex128: torch.onnx.TensorProtoDataType.COMPLEX128,
}


class Mark(torch.autograd.Function):
    """The mark node function.

    The function does nothing but inserts a mark node to ONNX model. The mark
    can ease the process of model partition.
    """

    @staticmethod
    def symbolic(g, x, dtype, shape, func, func_id, type, name, id, attrs):
        """Symbolic function for mmdeploy::Mark op."""
        n = g.op(
            "mmdeploy::Mark",
            x,
            dtype_i=TORCH_DTYPE_TO_ONNX[dtype],
            shape_i=shape,
            func_s=func,
            func_id_i=func_id,
            type_s=type,
            name_s=name,
            id_i=id,
            **attrs,
        )
        return n

    @staticmethod
    def forward(ctx, x, *args) -> torch.Tensor:
        """Run forward."""
        return x


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.core.optimizers.function_marker.Mark.symbolic"
)
def mark_symbolic(g, x, *args):
    """Rewrite symbolic of mark op."""
    ctx = FUNCTION_REWRITER.get_context()
    if cfg_apply_marks(ctx.cfg):
        return ctx.origin_func(g, x, *args)
    return x


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.core.optimizers.function_marker.Mark.forward"
)
def forward_of_mark(
    ctx, x, dtype, shape, func, func_id, type, name, id, attrs
) -> torch.Tensor:
    """Rewrite forward of mark op."""
    rewriter = FUNCTION_REWRITER.get_context()
    deploy_cfg = rewriter.cfg
    # save calib data
    apply_marks = cfg_apply_marks(deploy_cfg)
    create_calib = getattr(rewriter, "create_calib", False)
    if apply_marks and create_calib:
        partition_params = get_partition_config(deploy_cfg)
        assert partition_params is not None, "No partition config."
        partition_type = partition_params["type"]

        from juxtapose.mmdeploy.apis import get_predefined_partition_cfg

        partition_cfgs = get_predefined_partition_cfg(deploy_cfg, partition_type)
        assert hasattr(rewriter, "calib_file")

        for partition_id, partition_cfg in enumerate(partition_cfgs):
            start = partition_cfg["start"]
            if (f"{func}:{type}" not in start) and (
                f"{func}[{func_id}]:{type}" not in start
            ):
                continue

            input_name = name
            dynamic_axes = partition_cfg.get("dynamic_axes", None)
            if dynamic_axes is not None:
                input_name = name
            calib_file = rewriter.calib_file

            calib_data_group = calib_file["calib_data"]
            partition_name = f"partition{partition_id}"

            if partition_name not in calib_data_group:
                calib_data_group.create_group(partition_name)
            partition_group = calib_data_group[partition_name]

            if input_name not in partition_group:
                partition_group.create_group(input_name)
            input_data_group = partition_group[input_name]

            data_id = rewriter.data_id
            x_np = x.detach().cpu().numpy()
            input_data_group.create_dataset(
                str(data_id),
                shape=x_np.shape,
                compression="gzip",
                compression_opts=4,
                data=x_np,
            )

    return rewriter.origin_func(
        ctx, x, dtype, shape, func, func_id, type, name, id, attrs
    )


def mark_tensors(
    xs: Any,
    func: str,
    func_id: int,
    io_type: str,
    ctx: Any,
    attrs: Dict,
    is_inspecting: bool,
    level: int,
) -> tuple:
    """Add mark node recursively.

    Args:
        xs (Any): Input structure which contains tensor.
        func (str): Function name of the function which xs comes from.
        func_id (int): Function index of `func` in the model.
        io_type (str): The io type of xs, `input` or `output`.
        ctx (Any): The context instance.
        attrs (Dict): The extra attributes provided by mark decorator.
        is_inspecting (bool): The names of xs are inspected or not.
        level (int): The recursive level.

    Returns:
        Any: The same structure as xs, all tensor has been replaced with Mark.
    """
    visit = set()
    index = 0

    def impl(ys, prefix, level):
        nonlocal index
        old_index = index
        ret = ys
        prefix = () if level == 0 else prefix

        if isinstance(ys, torch.Tensor):
            if ys not in visit:
                visit.add(ys)
                root = ctx.names[ctx.index]
                name = ".".join(str(x) for x in (root, *prefix))
                ys_shape = tuple(int(s) for s in ys.shape)
                ret = Mark.apply(
                    ys, ys.dtype, ys_shape, func, func_id, io_type, name, index, attrs
                )
                index += 1
        elif isinstance(ys, list):
            ret = [impl(y, prefix + (i,), level + 1) for i, y in enumerate(ys)]
        elif isinstance(ys, tuple):
            ret = tuple(impl(y, prefix + (i,), level + 1) for i, y in enumerate(ys))
        elif isinstance(ys, dict):
            ret = {k: impl(v, prefix + (k,), level + 1) for k, v in ys.items()}

        if level == 0 and (is_inspecting or old_index != index):
            ctx.index += 1

        return ret

    return impl(xs, (), level)


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.core.optimizers.function_marker.mark_tensors", ir=IR.TORCHSCRIPT
)
def remove_mark__torchscript(xs: Any, *args, **kwargs):
    """Disable all marks for TorchScript backend.

    As the Node `mark` is not able to be traced, we just return original input
    for the function `mark_tensors`.

    Args:
        xs (Any): Input structure which contains tensor.
    """
    return xs


def mark(
    func_name: Optional[str] = None,
    inputs: Optional[Sequence[str]] = None,
    outputs: Optional[Sequence[str]] = None,
    **attrs,
) -> Callable:
    """The decorator used to add mark node.

    Mark node can be used to support model partition.

    Args:
        func_name (str): The name of the function where marks come from.
        inputs (Sequence[str]): The input names of the marks. The final name \
            might have suffix if inputs is list or dictionary.
        outputs (Sequence[str]): The output names of the marks. The final \
            name might have suffix if outputs is list or dictionary.

    Returns:
        Callable: The process of mark decorator.

    Examples:
        >>> from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
        >>> @FUNCTION_REWRITER.register_rewriter(
        >>>     func_name='mmdet.models.roi_heads.ConvFCBBoxHead.forward')
        >>> def forward_of_bbox_head(self, x):
        >>>     ctx = FUNCTION_REWRITER.get_context()
        >>>     @mark(
        >>>         'bbox_head_forward',
        >>>         inputs=['bbox_feats'],
        >>>         outputs=['cls_score', 'bbox_pred'])
        >>>     def _impl():
        >>>         return ctx.origin_func(self, x)
        >>>     return _impl()
    """
    MARK_FUNCTION_COUNT[func_name] = 0

    class Context:
        def __init__(self, names):
            self.names = names
            self.index = 0

    def decorator(f):
        func = func_name if func_name else f.__name__
        is_inspect = False
        if not inputs:
            input_names = list(inspect.signature(f).parameters.keys())
            is_inspect = True
        else:
            input_names = inputs
        output_names = outputs if outputs else func

        # args and retvals match corresponding names at level 0
        args_level, rets_level = -1, -1

        if isinstance(input_names, str):
            input_names = (input_names,)

        if isinstance(output_names, str):
            output_names = (output_names,)
            rets_level += 1

        def g(*args, **kwargs):
            func_id = MARK_FUNCTION_COUNT[func_name]
            MARK_FUNCTION_COUNT[func_name] += 1
            ctx = Context(input_names)
            args = mark_tensors(
                args, func, func_id, "input", ctx, attrs, is_inspect, args_level
            )

            rets = f(*args, **kwargs)

            ctx = Context(output_names)
            func_ret = mark_tensors(
                rets, func, func_id, "output", ctx, attrs, False, rets_level
            )
            return func_ret

        return g

    return decorator



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

import torch
from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import get_dynamic_axes


@FUNCTION_REWRITER.register_rewriter(func_name="torch.cat", backend="tensorrt")
def cat__tensorrt(tensors: Sequence[Tensor], *args, **kwargs) -> torch.Tensor:
    """Rewrite `cat` for TensorRT backend.

    cat in TensorRT does not support bool or uint8 type when input is dynamic.
    """
    ctx = FUNCTION_REWRITER.get_context()
    if get_dynamic_axes(ctx.cfg) is None:
        return ctx.origin_func(tensors, *args, **kwargs)
    if len(tensors) > 0 and (tensors[0].dtype in [torch.bool, torch.uint8]):
        original_dtype = tensors[0].dtype
        tensors = [i.to(torch.int32) for i in tensors]
        return ctx.origin_func(tensors, *args, **kwargs).to(original_dtype)
    return ctx.origin_func(tensors, *args, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
import torch.nn.functional as F
from mmengine.config import ConfigDict
from packaging import version
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops import ONNXNMSop, TRTBatchedNMSop


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.rtmdet_ins_head."
    "RTMDetInsHead.predict_by_feat"
)
def rtmdet_ins_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    kernel_preds: List[Tensor],
    mask_feat: Tensor,
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
):
    """Rewrite `predict_by_feat` of `RTMDet-Ins` for default backend.
    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.
    Args:
        ctx: Context that contains original meta information.
        cls_scores (list[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        batch_img_metas (list[dict], Optional): Batch image meta info.
            Defaults to None.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.
    Returns:
        tuple[Tensor, Tensor]: The first item is an (N, num_box, 5) tensor,
            where 5 represent (tl_x, tl_y, br_x, br_y, score), N is batch
            size and the score between 0 and 1. The shape of the second
            tensor in the tuple is (N, num_box), and each element
            represents the class label of the corresponding box.
    """
    assert len(cls_scores) == len(bbox_preds)
    device = cls_scores[0].device
    cfg = self.test_cfg if cfg is None else cfg
    batch_size = bbox_preds[0].shape[0]
    featmap_sizes = [cls_score.shape[2:] for cls_score in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, device=device, with_stride=True
    )

    flatten_cls_scores = [
        cls_score.permute(0, 2, 3, 1).reshape(batch_size, -1, self.cls_out_channels)
        for cls_score in cls_scores
    ]
    flatten_bbox_preds = [
        bbox_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, 4)
        for bbox_pred in bbox_preds
    ]
    flatten_kernel_preds = [
        kernel_pred.permute(0, 2, 3, 1).reshape(batch_size, -1, self.num_gen_params)
        for kernel_pred in kernel_preds
    ]
    flatten_cls_scores = torch.cat(flatten_cls_scores, dim=1).sigmoid()
    flatten_bbox_preds = torch.cat(flatten_bbox_preds, dim=1)
    flatten_kernel_preds = torch.cat(flatten_kernel_preds, dim=1)
    priors = torch.cat(mlvl_priors)
    tl_x = priors[..., 0] - flatten_bbox_preds[..., 0]
    tl_y = priors[..., 1] - flatten_bbox_preds[..., 1]
    br_x = priors[..., 0] + flatten_bbox_preds[..., 2]
    br_y = priors[..., 1] + flatten_bbox_preds[..., 3]
    bboxes = torch.stack([tl_x, tl_y, br_x, br_y], -1)
    # directly multiply score factor and feed to nms
    max_scores, _ = torch.max(flatten_cls_scores, 1)
    mask = max_scores >= cfg.score_thr
    scores = flatten_cls_scores.where(mask, flatten_cls_scores.new_zeros(1))

    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    mask_thr_binary = cfg.get("mask_thr_binary", 0.5)

    return _nms_with_mask_static(
        self,
        priors,
        bboxes,
        scores,
        flatten_kernel_preds,
        mask_feat,
        max_output_boxes_per_class,
        iou_threshold,
        score_threshold,
        pre_top_k,
        keep_top_k,
        mask_thr_binary,
    )


def _nms_with_mask_static(
    self,
    priors: Tensor,
    boxes: Tensor,
    scores: Tensor,
    kernels: Tensor,
    mask_feats: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    mask_thr_binary: float = 0.5,
):
    """Wrapper for `multiclass_nms` with ONNXRuntime.
    Args:
        ctx (ContextCaller): The context with additional information.
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.
    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    if version.parse(torch.__version__) < version.parse("1.13.0"):
        max_output_boxes_per_class = torch.LongTensor([max_output_boxes_per_class])
    iou_threshold = torch.tensor([iou_threshold], dtype=torch.float32)
    score_threshold = torch.tensor([score_threshold], dtype=torch.float32)

    # pre topk
    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.squeeze(0).topk(pre_top_k)
        boxes = boxes[:, topk_inds, :]
        scores = scores[:, topk_inds, :]
        kernels = kernels[:, topk_inds, :]
        priors = priors[topk_inds, :]

    scores = scores.permute(0, 2, 1)
    selected_indices = ONNXNMSop.apply(
        boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold
    )

    cls_inds = selected_indices[:, 1]
    box_inds = selected_indices[:, 2]

    scores = scores[:, cls_inds, box_inds].unsqueeze(2)
    boxes = boxes[:, box_inds, ...]
    kernels = kernels[:, box_inds, :]
    priors = priors[box_inds, :]
    dets = torch.cat([boxes, scores], dim=2)
    labels = cls_inds.unsqueeze(0)

    # pad
    dets = torch.cat((dets, dets.new_zeros((1, 1, 5))), 1)
    labels = torch.cat((labels, labels.new_zeros((1, 1))), 1)
    kernels = torch.cat((kernels, kernels.new_zeros(1, 1, kernels.shape[2])), 1)
    priors = torch.cat((priors, priors.new_zeros(1, 4)), 0)

    # topk or sort
    is_use_topk = keep_top_k > 0 and (
        torch.onnx.is_in_onnx_export() or keep_top_k < dets.shape[1]
    )
    if is_use_topk:
        _, topk_inds = dets[:, :, -1].topk(keep_top_k, dim=1)
    else:
        _, topk_inds = dets[:, :, -1].sort(dim=1, descending=True)
    topk_inds = topk_inds.squeeze(0)
    dets = dets[:, topk_inds, ...]
    labels = labels[:, topk_inds, ...]
    kernels = kernels[:, topk_inds, ...]
    priors = priors[topk_inds, ...]
    mask_logits = _mask_predict_by_feat_single(self, mask_feats, kernels[0], priors)
    stride = self.prior_generator.strides[0][0]
    mask_logits = F.interpolate(
        mask_logits.unsqueeze(0), scale_factor=stride, mode="bilinear"
    )
    masks = mask_logits.sigmoid()
    return dets, labels, masks


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.codebase.mmdet.models."
    "dense_heads.rtmdet_ins_head._nms_with_mask_static",
    backend="tensorrt",
)
def _nms_with_mask_static__tensorrt(
    self,
    priors: Tensor,
    boxes: Tensor,
    scores: Tensor,
    kernels: Tensor,
    mask_feats: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    mask_thr_binary: float = 0.5,
):
    """Wrapper for `multiclass_nms` with TensorRT.
    Args:
        ctx (ContextCaller): The context with additional information.
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.
    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    boxes = boxes if boxes.dim() == 4 else boxes.unsqueeze(2)
    keep_top_k = (
        max_output_boxes_per_class
        if keep_top_k < 0
        else min(max_output_boxes_per_class, keep_top_k)
    )
    dets, labels, inds = TRTBatchedNMSop.apply(
        boxes,
        scores,
        int(scores.shape[-1]),
        pre_top_k,
        keep_top_k,
        iou_threshold,
        score_threshold,
        -1,
        True,
    )
    # inds shape: (batch, n_boxes)
    # retain shape info
    batch_size = boxes.size(0)

    dets_shape = dets.shape
    label_shape = labels.shape
    dets = dets.reshape([batch_size, *dets_shape[1:]])
    labels = labels.reshape([batch_size, *label_shape[1:]])
    kernels = kernels[:, inds.reshape(-1), ...]
    priors = priors[inds.reshape(-1), ...]
    mask_logits = _mask_predict_by_feat_single(self, mask_feats, kernels[0], priors)
    stride = self.prior_generator.strides[0][0]
    mask_logits = F.interpolate(
        mask_logits.unsqueeze(0), scale_factor=stride, mode="bilinear"
    )
    masks = mask_logits.sigmoid()
    return dets, labels, masks


def _mask_predict_by_feat_single(self, mask_feat, kernels, priors):
    """decode mask with dynamic conv."""
    num_inst = priors.shape[0]
    h, w = mask_feat.size()[-2:]
    if num_inst < 1:
        return torch.empty(
            size=(num_inst, h, w), dtype=mask_feat.dtype, device=mask_feat.device
        )
    if len(mask_feat.shape) < 4:
        mask_feat.unsqueeze(0)
    coord = (
        self.prior_generator.single_level_grid_priors((h, w), level_idx=0)
        .reshape(1, -1, 2)
        .to(mask_feat.device)
    )
    num_inst = priors.shape[0]
    points = priors[:, :2].reshape(-1, 1, 2)
    strides = priors[:, 2:].reshape(-1, 1, 2)
    relative_coord = (points - coord).permute(0, 2, 1) / (
        strides[..., 0].reshape(-1, 1, 1) * 8
    )
    relative_coord = relative_coord.reshape(num_inst, 2, h, w)

    mask_feat = torch.cat([relative_coord, mask_feat.repeat(num_inst, 1, 1, 1)], dim=1)
    weights, biases = _parse_dynamic_params(self, kernels)

    n_layers = len(weights)
    x = mask_feat.flatten(2)
    for i, (weight, bias) in enumerate(zip(weights, biases)):
        # replace dynamic conv with bmm
        x = torch.bmm(weight, x)
        x = x + bias[:, :, None]
        if i < n_layers - 1:
            x = x.clamp_(min=0)
    x = x.reshape(num_inst, h, w)
    return x


def _parse_dynamic_params(self, flatten_kernels):
    """split kernel head prediction to conv weight and bias."""
    n_inst = flatten_kernels.size(0)
    n_layers = len(self.weight_nums)
    params_splits = list(
        torch.split_with_sizes(
            flatten_kernels, self.weight_nums + self.bias_nums, dim=1
        )
    )
    weight_splits = params_splits[:n_layers]
    bias_splits = params_splits[n_layers:]
    for idx in range(n_layers):
        if idx < n_layers - 1:
            weight_splits[idx] = weight_splits[idx].reshape(
                n_inst, self.dyconv_channels, -1
            )
        else:
            weight_splits[idx] = weight_splits[idx].reshape(n_inst, 1, -1)
    return weight_splits, bias_splits



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List

import mmengine

from juxtapose.mmdeploy.backend.openvino import ModelOptimizerOptions
from juxtapose.mmdeploy.utils import get_model_inputs
from juxtapose.mmdeploy.utils.config_utils import get_backend_config, get_ir_config


def update_input_names(
    input_info: Dict[str, List], input_names: List[str]
) -> Dict[str, List]:
    """Replaces the default input name in 'input_info' with the value from the
    deployment config, if they differ.

    Args:
        input_info (Dict[str, List]): Names and shapes of input.
        input_names (List[str]): Input names from the deployment config.

    Returns:
        Dict[str, List]: A dict that stores the names and shapes of input.
    """
    input_info_keys = set(input_info.keys())
    input_names = set(input_names)
    if input_info_keys != input_names:
        old_names = input_info_keys - input_names
        new_names = input_names - input_info_keys
        for new_key, old_key in zip(new_names, old_names):
            input_info[new_key] = input_info.pop(old_key)
    return input_info


def get_input_info_from_cfg(deploy_cfg: mmengine.Config) -> Dict[str, List]:
    """Get the input names and shapes from the configs for OpenVINO Model
    Optimizer.

    Args:
        deploy_cfg (mmengine.Config): Deployment config.

    Returns:
        Dict[str, List]: A dict that stores the names and shapes of input.
    """
    # The partition is not supported now. Set the id of model to 0.
    model_inputs = get_model_inputs(deploy_cfg)[0]
    input_info = model_inputs["opt_shapes"]
    ir_config = get_ir_config(deploy_cfg)
    if ir_config is not None:
        input_names = ir_config.get("input_names", None)
        if input_names:
            if not isinstance(input_info, Dict):
                input_info = dict(zip(input_names, input_info))
            input_info = update_input_names(input_info, input_names)
    return input_info


def get_mo_options_from_cfg(deploy_cfg: mmengine.Config) -> ModelOptimizerOptions:
    """Get additional parameters for the Model Optimizer from the deploy
    config.

    Args:
        deploy_cfg (mmengine.Config): Deployment config.

    Returns:
        ModelOptimizerOptions: A class that will contain additional arguments.
    """
    backend_config = get_backend_config(deploy_cfg)
    mo_options = backend_config.get("mo_options", None)
    mo_options = ModelOptimizerOptions(mo_options)
    return mo_options



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified from:
# https://github.com/pytorch/pytorch/blob/9ade03959392e5a90b74261012de1d806cab2253/torch/onnx/symbolic_opset9.py
from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@SYMBOLIC_REWRITER.register_symbolic(
    "hardsigmoid", is_pytorch=True, arg_descriptors=["v"]
)
def hardsigmoid__default(g, self):
    """Support export hardsigmoid This rewrite enable export hardsigmoid in
    torch<=1.8.2."""
    return g.op("HardSigmoid", self, alpha_f=1 / 6)



from pathlib import Path
from ultralytics import YOLO
import supervision as sv

from juxtapose.utils.downloads import safe_download

base = "pose"


class YOLOv8:
    """YOLOv8 model (s, m, l) to detect only person (class 0)"""

    def __init__(self, type: str = "m", device: str = "cpu", conf_thres: float = 0.3):
        # onnx_file = Path(f"model/rtmdet-{type}.onnx")
        # if not onnx_file.exists():
        #     safe_download(
        #         f"https://huggingface.co/ziq/rtm/resolve/main/rtmdet-{type}.onnx",
        #         file=f"rtmdet-{type}",
        #         dir=Path(f"model/"),
        #     )
        self.model = YOLO(f"yolov8{type}.pt")
        self.model.to(device)
        self.conf_thres = conf_thres

    def __call__(self, im):
        """Return List of xyxy coordinates based on im frame (cv2.imread)
        im -> (h, w, c)
        return -> [[x, y, x, y], [x, y, x, y], ...] -> two persons detected
        """
        result = self.model(im, verbose=False, conf=self.conf_thres)[0].boxes
        result = result[result.cls == 0].cpu().numpy()
        result = sv.Detections(
            xyxy=result.xyxy,
            confidence=result.conf,
        )
        result.labels = [0 for _ in range(len(result.xyxy))]

        return result



from transformers import AutoTokenizer, BertModel, BertTokenizer, RobertaModel, RobertaTokenizerFast
import os

def get_tokenlizer(text_encoder_type):
    if not isinstance(text_encoder_type, str):
        # print("text_encoder_type is not a str")
        if hasattr(text_encoder_type, "text_encoder_type"):
            text_encoder_type = text_encoder_type.text_encoder_type
        elif text_encoder_type.get("text_encoder_type", False):
            text_encoder_type = text_encoder_type.get("text_encoder_type")
        elif os.path.isdir(text_encoder_type) and os.path.exists(text_encoder_type):
            pass
        else:
            raise ValueError(
                "Unknown type of text_encoder_type: {}".format(type(text_encoder_type))
            )
    print("final text_encoder_type: {}".format(text_encoder_type))

    tokenizer = AutoTokenizer.from_pretrained(text_encoder_type)
    return tokenizer


def get_pretrained_language_model(text_encoder_type):
    if text_encoder_type == "bert-base-uncased" or (os.path.isdir(text_encoder_type) and os.path.exists(text_encoder_type)):
        return BertModel.from_pretrained(text_encoder_type)
    if text_encoder_type == "roberta-base":
        return RobertaModel.from_pretrained(text_encoder_type)

    raise ValueError("Unknown text_encoder_type {}".format(text_encoder_type))



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.onnx.symbolic_helper as sym_help

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import get_ir_config


@SYMBOLIC_REWRITER.register_symbolic("squeeze", is_pytorch=True)
def squeeze__default(g, self, dim=None):
    """Register default symbolic function for `squeeze`.

    squeeze might be exported with IF node in ONNX, which is not supported in
    lots of backend.
    """
    if dim is None:
        dims = []
        for i, size in enumerate(self.type().sizes()):
            if size == 1:
                dims.append(i)
    else:
        dims = [sym_help._get_const(dim, "i", "dim")]

    ctx = SYMBOLIC_REWRITER.get_context("squeeze")
    if get_ir_config(ctx.cfg).get("opset_version", 11) >= 13:
        axes = g.op("Constant", value_t=torch.tensor(dims, dtype=torch.long))
        return g.op("Squeeze", self, axes)

    return g.op("Squeeze", self, axes_i=dims)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

import torch
import torch.nn.functional as F
from mmengine import ConfigDict
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_backend


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads." "mask_heads.fcn_mask_head.FCNMaskHead.predict_by_feat"
)
def fcn_mask_head__predict_by_feat(
    self,
    mask_preds: Tuple[Tensor],
    results_list: List[Tensor],
    batch_img_metas: List[dict],
    rcnn_test_cfg: ConfigDict,
    rescale: bool = False,
    activate_map: bool = False,
) -> List[Tensor]:
    """Transform a batch of output features extracted from the head into mask
    results.

    Args:
        mask_preds (tuple[Tensor]): Tuple of predicted foreground masks,
            each has shape (n, num_classes, h, w).
        results_list (list[Tensor]): Detection results of
            each image.
        batch_img_metas (list[dict]): List of image information.
        rcnn_test_cfg (obj:`ConfigDict`): `test_cfg` of Bbox Head.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        activate_map (book): Whether get results with augmentations test.
            If True, the `mask_preds` will not process with sigmoid.
            Defaults to False.

    Returns:
        list[Tensor]: Detection results of each image
        after the post process. Each item usually contains following keys.

            - dets (Tensor): Classification scores, has a shape
                (num_instance, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
            - masks (Tensor): Has a shape (num_instances, H, W).
    """
    ctx = FUNCTION_REWRITER.get_context()
    ori_shape = batch_img_metas[0]["img_shape"]
    dets, det_labels = results_list
    dets = dets.view(-1, 5)
    det_labels = det_labels.view(-1)
    backend = get_backend(ctx.cfg)
    mask_preds = mask_preds.sigmoid()
    bboxes = dets[:, :4]
    labels = det_labels
    threshold = rcnn_test_cfg.mask_thr_binary
    if not self.class_agnostic:
        box_inds = torch.arange(mask_preds.shape[0], device=bboxes.device)
        mask_pred = mask_preds[box_inds, labels][:, None]

    # grid sample is not supported by most engine
    # so we add a flag to disable it.
    mmdet_params = get_post_processing_params(ctx.cfg)
    export_postprocess_mask = mmdet_params.get("export_postprocess_mask", False)
    if not export_postprocess_mask:
        return mask_pred

    masks, _ = _do_paste_mask(
        mask_pred, bboxes, ori_shape[0], ori_shape[1], skip_empty=False
    )
    if backend == Backend.TENSORRT:
        return masks
    if threshold >= 0:
        masks = (masks >= threshold).to(dtype=torch.bool)
    return masks


def _do_paste_mask(masks, boxes, img_h, img_w, skip_empty=True):
    """Paste instance masks according to boxes.

    This implementation is modified from
    https://github.com/facebookresearch/detectron2/

    Args:
        masks (Tensor): N, 1, H, W
        boxes (Tensor): N, 4
        img_h (int): Height of the image to be pasted.
        img_w (int): Width of the image to be pasted.
        skip_empty (bool): Only paste masks within the region that
            tightly bound all boxes, and returns the results this region only.
            An important optimization for CPU.

    Returns:
        tuple: (Tensor, tuple). The first item is mask tensor, the second one
            is the slice object.
        If skip_empty == False, the whole image will be pasted. It will
            return a mask of shape (N, img_h, img_w) and an empty tuple.
        If skip_empty == True, only area around the mask will be pasted.
            A mask of shape (N, h', w') and its start and end coordinates
            in the original image will be returned.
    """
    # On GPU, paste all masks together (up to chunk size)
    # by using the entire image to sample the masks
    # Compared to pasting them one by one,
    # this has more operations but is faster on COCO-scale dataset.
    device = masks.device
    if skip_empty:
        x0_int, y0_int = torch.clamp(boxes.min(dim=0).values.floor()[:2] - 1, min=0).to(
            dtype=torch.int32
        )
        x1_int = torch.clamp(boxes[:, 2].max().ceil() + 1, max=img_w).to(
            dtype=torch.int32
        )
        y1_int = torch.clamp(boxes[:, 3].max().ceil() + 1, max=img_h).to(
            dtype=torch.int32
        )
    else:
        x0_int, y0_int = 0, 0
        x1_int, y1_int = img_w, img_h
    x0, y0, x1, y1 = torch.split(boxes, 1, dim=1)  # each is Nx1

    N = masks.shape[0]

    img_y = torch.arange(y0_int, y1_int, device=device).to(torch.float32) + 0.5
    img_x = torch.arange(x0_int, x1_int, device=device).to(torch.float32) + 0.5
    img_y = (img_y - y0) / (y1 - y0) * 2 - 1
    img_x = (img_x - x0) / (x1 - x0) * 2 - 1
    gx = img_x[:, None, :].expand(N, img_y.size(1), img_x.size(1))
    gy = img_y[:, :, None].expand(N, img_y.size(1), img_x.size(1))
    grid = torch.stack([gx, gy], dim=3)

    img_masks = F.grid_sample(masks.to(dtype=torch.float32), grid, align_corners=False)

    if skip_empty:
        return img_masks[:, 0], (slice(y0_int, y1_int), slice(x0_int, x1_int))
    else:
        return img_masks[:, 0], ()



_base_ = "./rtmdet-l.py"

model = dict(
    backbone=dict(deepen_factor=0.67, widen_factor=0.75),
    neck=dict(in_channels=[192, 384, 768], out_channels=192, num_csp_blocks=2),
    bbox_head=dict(in_channels=192, feat_channels=192),
)



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Any, Optional, Union

import mmengine

from juxtapose.mmdeploy.apis.core.pipeline_manager import PIPELINE_MANAGER, no_mp


@PIPELINE_MANAGER.register_pipeline()
def torch2torchscript(
    img: Any,
    work_dir: str,
    save_file: str,
    deploy_cfg: Union[str, mmengine.Config],
    model_cfg: Union[str, mmengine.Config],
    model_checkpoint: Optional[str] = None,
    device: str = "cuda:0",
):
    """Convert PyTorch model to torchscript model.

    Args:
        img (str | np.ndarray | torch.Tensor): Input image used to assist
            converting model.
        work_dir (str): A working directory to save files.
        save_file (str): Filename to save torchscript model.
        deploy_cfg (str | mmengine.Config): Deployment config file or
            Config object.
        model_cfg (str | mmengine.Config): Model config file or Config object.
        model_checkpoint (str): A checkpoint path of PyTorch model,
            defaults to `None`.
        device (str): A string specifying device type, defaults to 'cuda:0'.
    """
    import torch

    from juxtapose.mmdeploy.utils import get_backend, get_input_shape, load_config
    from .torch_jit import trace

    # load deploy_cfg if necessary
    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)
    mmengine.mkdir_or_exist(osp.abspath(work_dir))

    input_shape = get_input_shape(deploy_cfg)

    from juxtapose.mmdeploy.apis import build_task_processor

    task_processor = build_task_processor(model_cfg, deploy_cfg, device)
    torch_model = task_processor.build_pytorch_model(model_checkpoint)
    data, model_inputs = task_processor.create_input(
        img,
        input_shape,
        data_preprocessor=getattr(torch_model, "data_preprocessor", None),
    )

    data_samples = data["data_samples"]
    input_metas = {"data_samples": data_samples, "mode": "predict"}

    if not isinstance(model_inputs, torch.Tensor) and len(model_inputs) == 1:
        model_inputs = model_inputs[0]

    context_info = dict(deploy_cfg=deploy_cfg)
    backend = get_backend(deploy_cfg).value
    output_prefix = osp.join(work_dir, osp.splitext(save_file)[0])
    if model_inputs.device != device:
        model_inputs = model_inputs.to(device)
    with no_mp():
        trace(
            torch_model,
            model_inputs,
            output_path_prefix=output_prefix,
            backend=backend,
            input_metas=input_metas,
            context_info=context_info,
            check_trace=False,
        )
